# SPDX-FileCopyrightText: 2023 Karlsruher Institut für Technologie
#
# SPDX-License-Identifier: CC0-1.0


from pystac.extensions.datacube import (
    DatacubeExtension,
    Dimension,
    DimensionType,
    Variable,
    VariableType,
)

from ..logger import Logger


class Datacube(object):
    """
    This class is responsible for adding the datacube extension to the STAC Item.
    Args:
        item (pystac.Item): The STAC Item to be extended.
        harvesting_vars (dict): The dictionary of the variables and dimensions of the dataset.
        logger_properties (dict): The dictionary of the logger properties.
    """

    def item_extension(
        self, item, harvesting_vars, logger_properties: dict = dict()
    ):
        self.logger_properties = logger_properties
        variables = (
            {}
        )  # variables dictionary for gathering the Variable objects
        dimensions = (
            {}
        )  # dimensions dictionary for gathering the Dimension objects
        cube = DatacubeExtension.ext(item, add_if_missing=True)
        if harvesting_vars.get("variable_dimensions") is not None:
            variable_dimensions = [
                elem.replace(" ", ",").split(",") if " " in elem else [elem]
                for elem in harvesting_vars["variable_dimensions"]
            ]
        else:
            self.logger_properties["logger_level"] = "WARNING"
            self.logger_properties[
                "logger_msg"
            ] = "The variable's dimensions list is not involved in the output dictionary. Check your dataset in Thredds or correct keys in your `tag_config.json` file to produce new lists. Datacube extension is not added to the STAC Item."
            Logger(self.logger_properties)
            return
        if len(harvesting_vars["variable_ids"]) != len(variable_dimensions):
            # Another solution for this is to index each output element and find the None values to decide about them later
            if harvesting_vars["variable_ids"] is None:
                harvesting_vars["variable_ids"] = []
            self.logger_properties["logger_level"] = "WARNING"
            self.logger_properties[
                "logger_msg"
            ] = "The length of the variable's IDs list and the variable's dimensions list are not equal. Check your dataset in Thredds or correct keys in your `tag_config.json` file to produce new lists. Datacube extension is not added to the STAC Item."
            Logger(self.logger_properties)
            return
        else:
            for i, v in enumerate(harvesting_vars["variable_ids"]):
                variable_dict = dict()
                variable_dict["dimensions"] = variable_dimensions[i]
                variable_dict["type"] = VariableType.DATA.value
                if (
                    harvesting_vars["variable_description"] is not None
                    and len(harvesting_vars["variable_ids"])
                    == len(harvesting_vars["variable_description"])
                    and len(harvesting_vars["variable_description"])
                    != harvesting_vars["variable_description"].count(None)
                ):
                    if harvesting_vars["variable_description"][i] is not None:
                        variable_dict["description"] = harvesting_vars[
                            "variable_description"
                        ][i]
                variable_dict["dimensions"] = variable_dimensions[i]
                if (
                    harvesting_vars["variable_unit"] is not None
                    and len(harvesting_vars["variable_unit"])
                    == len(variable_dimensions)
                    and len(harvesting_vars["variable_unit"])
                    != harvesting_vars["variable_unit"].count(None)
                ):
                    if harvesting_vars["variable_unit"][i] is not None:
                        variable_dict["units"] = harvesting_vars[
                            "variable_unit"
                        ][i]
                variables[harvesting_vars["variable_ids"][i]] = Variable(
                    variable_dict
                )

        # Horizontal Spatial Dimension
        list_of_required_keys = [
            "horizontal_ids_lat",
            "horizontal_ids_lon",
            "horizontal_axis_x",
            "horizontal_axis_y",
            "horizontal_extent_lon_min",
            "horizontal_extent_lon_max",
            "horizontal_extent_lat_min",
            "horizontal_extent_lat_max",
        ]

        if all(
            harvesting_vars.get(key) is not None
            for key in list_of_required_keys
        ):
            horizontal_dict = dict()
            horizontal_dict = {
                "type": DimensionType.SPATIAL.value,
                "axis": harvesting_vars["horizontal_axis_x"],
                "description": harvesting_vars["horizontal_description_lon"],
                "extent": [
                    float(
                        harvesting_vars["horizontal_extent_lon_min"].replace(
                            ",", "."
                        )
                    ),
                    float(
                        harvesting_vars["horizontal_extent_lon_max"].replace(
                            ",", "."
                        )
                    ),
                ],
                "reference_system": harvesting_vars[
                    "horizontal_reference_system"
                ],
            }
            dimensions[harvesting_vars["horizontal_ids_lon"]] = Dimension(
                horizontal_dict
            )
            horizontal_dict = dict()
            horizontal_dict = {
                "type": DimensionType.SPATIAL.value,
                "axis": harvesting_vars["horizontal_axis_y"],
                "description": harvesting_vars["horizontal_description_lat"],
                "extent": [
                    float(
                        harvesting_vars["horizontal_extent_lat_min"].replace(
                            ",", "."
                        )
                    ),
                    float(
                        harvesting_vars["horizontal_extent_lat_max"].replace(
                            ",", "."
                        )
                    ),
                ],
                "reference_system": harvesting_vars[
                    "horizontal_reference_system"
                ],
            }
            dimensions[harvesting_vars["horizontal_ids_lat"]] = Dimension(
                horizontal_dict
            )
        else:
            logger_string_output = ""
            logger_properties["logger_level"] = "WARNING"
            for key in list_of_required_keys:
                if harvesting_vars.get(key) is None:
                    logger_string_output = (
                        logger_string_output
                        + f"{key}: {harvesting_vars.get(key)} \n"
                    )
            logger_properties[
                "logger_msg"
            ] = f"Required horizontal attributes aren't involved in the output dictionary. Check your dataset in Thredds or check the following lists and reconfig your `tag_config.json` file. {logger_string_output}"
            Logger(logger_properties)
            return

        # Vertical Spatial Dimension
        list_of_required_keys = [
            "vertical_id",
            "vertical_axis",
            "vertical_extent_upper",
            "vertical_extent_lower",
        ]
        if all(
            harvesting_vars.get(key) is not None and harvesting_vars[key] != []
            for key in list_of_required_keys
        ):
            vertical_dict = dict()
            vertical_dict = {
                "type": DimensionType.SPATIAL.value,
                "axis": harvesting_vars["vertical_axis"],
                "description": harvesting_vars["vertical_description"],
                "extent": [
                    float(
                        harvesting_vars["vertical_extent_lower"].replace(
                            ",", "."
                        )
                    ),
                    float(
                        harvesting_vars["vertical_extent_upper"].replace(
                            ",", "."
                        )
                    ),
                ],
            }
            dimensions[harvesting_vars["vertical_id"]] = Dimension(
                vertical_dict
            )
        else:
            logger_string_output = ""
            # logger_properties["logger_level"] = "WARNING"
            for key in list_of_required_keys:
                if harvesting_vars.get(key) is None:
                    logger_string_output = (
                        logger_string_output
                        + f"{key}: {harvesting_vars.get(key)} \n"
                    )
            # logger_properties[
            #    "logger_msg"
            # ] = f"Required vertical attributes aren't involved in the output dictionary. Check your dataset in Thredds or check the following lists and reconfig your `tag_config.json` file. {logger_string_output}"
            # Logger(logger_properties)

        # Temporal Dimension
        list_of_required_keys = [
            "temporal_id",
            "temporal_extent_start_datetime",
            "temporal_extent_end_datetime",
        ]
        if all(
            harvesting_vars.get(key) is not None
            for key in list_of_required_keys
        ):
            temporal_dict = dict()
            temporal_dict = {
                "type": DimensionType.TEMPORAL.value,
                "description": harvesting_vars["temporal_description"],
                "extent": [
                    harvesting_vars["temporal_extent_start_datetime"],
                    harvesting_vars["temporal_extent_end_datetime"],
                ],
            }
            dimensions[harvesting_vars["temporal_id"]] = Dimension(
                temporal_dict
            )
        else:
            logger_string_output = ""
            logger_properties["logger_level"] = "WARNING"
            for key in list_of_required_keys:
                if harvesting_vars.get(key) is None:
                    logger_string_output = (
                        logger_string_output
                        + f"{key}: {harvesting_vars.get(key)} \n"
                    )

            logger_properties[
                "logger_msg"
            ] = f"Required attributes aren't involved in the output dictionary. Check your dataset in Thredds or check the following lists and reconfig your `tag_config.json` file. \n{logger_string_output}"
            Logger(logger_properties)
            return

        # Additional Dimensions
        anotherlist = [
            "x",
            "y",
            "time",
            "longitude",
            "latitude",
            "t",
            "z",
            "height",
        ]
        for id in harvesting_vars["additional_ids"]:
            if id.lower() not in [
                i.lower() for i in list(dimensions.keys()) + anotherlist
            ]:
                additional_dict = dict()
                additional_dict = {
                    "type": harvesting_vars["additional_type"],
                    "extent": [None, None],
                }
                dimensions[id] = Dimension(additional_dict)
        cube.apply(dimensions=dimensions, variables=variables)



# SPDX-FileCopyrightText: 2023 Karlsruher Institut für Technologie
#
# SPDX-License-Identifier: CC0-1.0

import os
from urllib import parse as urlparse
from urllib.request import (
    HTTPBasicAuthHandler,
    HTTPPasswordMgrWithDefaultRealm,
    build_opener,
)

import requests


def replacement_func(url):
    """A function for making a an id from catalog URL for collections"""
    split_arr = urlparse.urlsplit(url)
    trans_dict_path = {"%": "", "/catalog.xml": "", "/thredds/": "", "/": " "}
    path_ = split_arr.path
    query_ = split_arr.query
    for k, v in trans_dict_path.items():
        temp = path_.split(k)
        path_ = v.join(temp)
    if split_arr.query != "":
        trans_dict_query = {"=": " ", "?": " ", "/": " "}
        for k, v in trans_dict_query.items():
            temp = query_.split(k)
            query_ = v.join(temp)
        replaced_url = path_.replace(".xml", "") + " " + query_
        trans_dict_replaced = {".nc": "", "-": " ", "_": " "}
        for k, v in trans_dict_replaced.items():
            temp = replaced_url.split(k)
            replaced_url = v.join(temp)
    else:
        replaced_url = path_
        trans_dict_replaced = {".xml": "", ".nc": "", "-": " ", "_": " "}
        for k, v in trans_dict_replaced.items():
            temp = replaced_url.split(k)
            replaced_url = v.join(temp)
    replaced_url = replaced_url.replace(".", " ")
    return replaced_url.title()


def replacement_func_collection_item_id(url):
    """A function for making a an id from catalog URL for collections"""
    split_arr = urlparse.urlsplit(url)
    trans_dict_path = {"%": "", "/catalog.xml": "", "/thredds/": "", "/": " "}
    path_ = split_arr.path
    query_ = split_arr.query
    for k, v in trans_dict_path.items():
        temp = path_.split(k)
        path_ = v.join(temp)
    if split_arr.query != "":
        trans_dict_query = {"=": " ", "?": " ", "/": " "}
        for k, v in trans_dict_query.items():
            temp = query_.split(k)
            query_ = v.join(temp)
        replaced_url = path_.replace(".xml", "") + " " + query_
        trans_dict_replaced = {".nc": "", "-": " ", "_": " "}
        for k, v in trans_dict_replaced.items():
            temp = replaced_url.split(k)
            replaced_url = v.join(temp)
    else:
        replaced_url = path_
        trans_dict_replaced = {".xml": "", ".nc": "", "-": " ", "_": " "}
        for k, v in trans_dict_replaced.items():
            temp = replaced_url.split(k)
            replaced_url = v.join(temp)
    replaced_url = replaced_url.lower()
    replaced_url = replaced_url.replace(" ", "_")
    replaced_url = replaced_url.replace(".", "_")
    return replaced_url


def html2xml(url):
    """A function for making a an xml URL from html URL"""
    u = urlparse.urlsplit(url)
    path, extension = os.path.splitext(u.path)
    if extension == ".html":
        u = urlparse.urlsplit(url.replace(".html", ".xml"))
    return u.geturl()


def xml2html(url):
    """A function for making a an html URL from xml URL"""
    u = urlparse.urlsplit(url)
    path, extension = os.path.splitext(u.path)
    if extension == ".xml":
        u = urlparse.urlsplit(url.replace(".xml", ".html"))
    return u.geturl()


def get_xml(url, request_properties):
    """A function for getting XML content from url"""

    try:
        if request_properties is not None and request_properties != {}:
            xml_url = requests.get(
                url,
                None,
                auth=request_properties["auth"],
                verify=request_properties["verify"],
                timeout=request_properties["timeout"],
            )
            xml = xml_url.text.encode("utf-8")
            return xml
        else:
            xml_url = requests.get(url)
            xml = xml_url.text.encode("utf-8")
            return xml
    except BaseException:
        pass
        return None


def references_urls(url, additional):
    split_arr = urlparse.urlsplit(url)
    common_url = str(split_arr.scheme) + "://" + str(split_arr.netloc)
    without_catalog_xml = urlparse.urljoin(
        common_url, os.path.split(split_arr.path)[0]
    )

    if not additional:
        final_url = url
    elif additional[:4] == "http":
        # finding http or https
        final_url = additional
    elif additional[0] == "/":
        # Absolute paths
        final_url = urlparse.urljoin(common_url, additional)
    else:
        # Relative paths.
        final_url = without_catalog_xml + "/" + additional
    return final_url


def xml_processing(catalog, request_properties):
    """A function for getting out XML details of a catalog URL"""
    catalog_xml = html2xml(catalog)
    catalog_id = replacement_func(catalog_xml)
    xml_final = get_xml(catalog_xml, request_properties)
    return catalog_xml, catalog_id, xml_final


def xml_tag_name_ncml(input_xml, var_name):
    """A function for finding the tag names in NcML XML files"""

    # A list for recognizign the exceptions.
    # This list contains variable's name with same `input_xml.tag` and different `var_name`s
    exception_list = ["var_lists", "var_dims", "var_descs", "keyword"]

    if var_name in exception_list:
        return input_xml.tag + "_" + var_name
    else:
        return str(input_xml.get("name")) + "_" + var_name


def xml_tag_finder(input_xml, web_service, var_name):
    """A function for finding the tag names in all TDS webservices"""

    tag_finder_dict = {
        "iso": input_xml.tag + "_" + var_name,
        "ncml": xml_tag_name_ncml(input_xml, var_name),
        "wms": input_xml.tag + "_" + var_name,
    }

    return tag_finder_dict.get(
        web_service,
    )


def validate_catalog_url(url, requests_properties):
    """A function for validating the catalog URL"""
    try:
        if requests_properties is not None and requests_properties != {}:
            xml_url = requests.get(
                url,
                None,
                auth=requests_properties["auth"],
                verify=requests_properties["verify"],
                timeout=requests_properties["timeout"],
            )
            if xml_url.status_code == 200:
                return True
            else:
                return False
        else:
            xml_url = requests.get(url)
            if xml_url.status_code == 200:
                return True
            else:
                return False
    except BaseException:
        return False


def merge_bboxes(bbox1, bbox2):
    x1_1, y1_1, x2_1, y2_1 = bbox1
    x1_2, y1_2, x2_2, y2_2 = bbox2
    new_x1 = min(x1_1, x1_2)
    new_y1 = min(y1_1, y1_2)
    new_x2 = max(x2_1, x2_2)
    new_y2 = max(y2_1, y2_2)
    return (new_x1, new_y1, new_x2, new_y2)


def merge_intervals(interval1, interval2):
    start1, end1 = interval1
    start2, end2 = interval2
    merged_start = min(start1, start2)
    merged_end = max(end1, end2)
    return (merged_start, merged_end)


def opener_module(service_url, requests_properties):
    # Set your URL, username, and password

    username = requests_properties["auth"][0]
    password = requests_properties["auth"][1]
    # Create a password manager
    password_mgr = HTTPPasswordMgrWithDefaultRealm()
    password_mgr.add_password(None, service_url, username, password)

    # Create an opener that will replace the default urlopen method on further calls
    handler = HTTPBasicAuthHandler(password_mgr)
    opener = build_opener(handler)
    opener.open(service_url)



import importlib
import inspect

# SPDX-FileCopyrightText: 2023 Karlsruher Institut für Technologie
#
# SPDX-License-Identifier: CC0-1.0
import os
import sys

# import traceback
from datetime import datetime
from importlib.machinery import SourceFileLoader
from importlib.util import module_from_spec, spec_from_loader
from typing import Union

import pystac
from shapely import geometry

from . import assets, logger, utils
from .analysers.existence_validator import ExistenceValidator
from .extensions import common_metadata, datacube
from .extensions.extra_metadata import ExtraMetadata
from .statics import constants


class STACCreator(object):
    """
    A class for creating STAC catalog, -Collections
    and its -Items from TDS datasets catalogs.
    """

    def STACCatalog(
        self,
        url: str,
        stac_id: str,
        stac_title: Union[str, None],
        stac_desc: Union[str, None],
        stac_dir: str,
        stac_existence: bool = False,
        logger_properties: dict = dict(),
        requests_properties: dict = dict(),
    ):
        """
        A function for creating STAC catalog from TDS dataset catalog.

        Args:
            url: The URL of the TDS catalog.
            stac_id: The ID of the STAC catalog.
            stac_title: The title of the STAC catalog.
            stac_desc: The description of the STAC catalog.
            stac_dir: The directory of saving the STAC catalog.
            stac_existence: If it is True, it means that the STAC catalog
                already exists in the directory and for the harvesting, there
                is no need to create a new STAC-Catalog and import new collections
                In the existed STAC-Catalog. False by default.
            logger_properties: The properties of the logger. For more information please check the :class:`~tds2stac.logger.Logger` class.
            requests_properties: The properties of the requests. For more information please check the :class:`~tds2stac.TDS2STAC.requests_properties` class.
        """
        # using 'xml_processing' we get the catalog URL with
        # XML extension and catalog id and XML content of TDS catalog.
        catalog: dict = dict()

        if logger_properties is not None:
            self.logger_properties = logger_properties
        xml_url_catalog, id_catalog, xml = utils.xml_processing(
            url, requests_properties
        )
        if stac_desc is None:
            stac_desc = "This is a STAC catalog created by tds2stac"

        # In the following if condition we are going to create a new STAC catalog or use the existed one.
        if stac_existence is True:
            if stac_dir is None:
                self.logger_properties["logger_level"] = "WARNING"
                self.logger_properties[
                    "logger_msg"
                ] = "You have turned on the `stac_existence`, so please provide the directory of the existed STAC catalog"
                logger.Logger(self.logger_properties)
                return
            else:
                if ExistenceValidator(stac_dir).existence is True:
                    self.logger_properties["logger_level"] = "INFO"
                    self.logger_properties[
                        "logger_msg"
                    ] = "The STAC catalog already exists in the directory"
                    logger.Logger(self.logger_properties)
                    id_catalog = pystac.Catalog.from_file(
                        stac_dir + "/catalog.json"
                    ).id
                    catalog[id_catalog] = pystac.Catalog.from_file(
                        stac_dir + "/catalog.json"
                    )
                else:
                    self.logger_properties["logger_level"] = "INFO"
                    self.logger_properties[
                        "logger_msg"
                    ] = "The STAC catalog does not exist in the directory"
                    logger.Logger(self.logger_properties)
                    id_catalog = id_catalog + " Catalog"
                    catalog[id_catalog] = pystac.Catalog(
                        id=stac_id,
                        title=stac_title,
                        description="["
                        + stac_desc
                        + "]("
                        + utils.xml2html(xml_url_catalog)
                        + ")",
                    )
        else:
            self.logger_properties["logger_level"] = "INFO"
            self.logger_properties[
                "logger_msg"
            ] = "It creates a new catalog in the directory"
            logger.Logger(self.logger_properties)
            id_catalog = id_catalog + " Catalog"
            catalog[id_catalog] = pystac.Catalog(
                id=stac_id,
                title=stac_title,
                description="["
                + stac_desc
                + "]("
                + utils.xml2html(xml_url_catalog)
                + ")",
            )
        return catalog[id_catalog]

    def STACCollection(
        self,
        catalog: pystac.Catalog,
        collection_id: str,
        collection_title: str,
        collection_description: str,
        # collection_scientific: bool = False,
        stac_existence_collection: bool = False,
        logger_properties: dict = dict(),
        extra_metadata: dict = dict(),
    ):
        """
        This is a function for creating STAC collection
        from harvested information from TDS dataset catalog.
        This function returns a dictionary with two keys:
            1. collection: The STAC collection

            2. existed_items_id_list: The list of the items that already
            exist in the STAC collection and it is going to be used for
            the harvesting process.

        Args:
            catalog: The STAC catalog.
            collection_id: The ID of the STAC collection.
            collection_title: The title of the STAC collection.
            collection_description: The description of the STAC collection.
            collection_scientific: The scientific extension of the STAC collection.
            stac_existence_collection: If it is True, it means that the STAC collection
                already exists in the catalog and for the harvesting, there
                is no need to create a new STAC-Collection and import new items
                In the existed STAC-Collection. False by default.
            logger_properties: The properties of the logger. For more information please check the :class:`~tds2stac.logger.Logger` class.
        """
        collection: dict = dict()
        collection["collection_bbox_existed"] = None
        collection["collection_interval_time_final_existed"] = None
        existed_items_id_list = []
        if logger_properties is not None:
            self.logger_properties = logger_properties
        if stac_existence_collection is True:
            existed_collections_id_list = []
            existed_collections_id_list = [
                existence_collection.id
                for existence_collection in list(catalog.get_collections())
            ]
            if (
                collection is not None
                and collection_id in existed_collections_id_list
            ):
                collection[collection_id] = catalog.get_child(collection_id)
                collection["collection_bbox_existed"] = collection[
                    collection_id
                ].extent.spatial.bboxes[0]
                collection[
                    "collection_interval_time_final_existed"
                ] = collection[collection_id].extent.temporal.intervals[0]
                existed_items_id_list = [
                    existed_item.id
                    for existed_item in list(
                        collection[collection_id].get_items()
                    )
                ]
            else:
                # Defining a None Spatial and Temporal extent for the collection
                collection[collection_id] = pystac.Collection(
                    id=collection_id,
                    title=collection_title,
                    extent=pystac.Extent(
                        spatial=pystac.SpatialExtent(bboxes=[0.0, 0.0]),
                        temporal=pystac.TemporalExtent(
                            intervals=[[datetime.utcnow(), datetime.utcnow()]]
                        ),
                    ),
                    description=collection_description,
                )
                if extra_metadata is not None:
                    if extra_metadata.get("extra_metadata"):
                        ExtraMetadata(
                            logger_properties=self.logger_properties
                        ).collection(
                            collection=collection[collection_id],
                            extra_metadata=extra_metadata,
                        )
                    else:
                        self.logger_properties["logger_level"] = "INFO"
                        self.logger_properties[
                            "logger_msg"
                        ] = "The `extra_metadata` is not activated. So, it does not add any extra metadata to the STAC collection."
                        # logger.Logger(self.logger_properties)

                catalog.add_child(collection[collection_id])

        else:
            # When the STAC collection does not exist in the catalog
            # Instead of None value for Spatial and Temporal extent, we define a default value
            # as a list of [0.0, 0.0] for Spatial extent and [[datetime.utcnow(), datetime.utcnow()]] for Temporal extent
            collection[collection_id] = pystac.Collection(
                id=collection_id,
                title=collection_title,
                extent=pystac.Extent(
                    spatial=pystac.SpatialExtent(bboxes=[0.0, 0.0]),
                    temporal=pystac.TemporalExtent(
                        intervals=[[datetime.utcnow(), datetime.utcnow()]]
                    ),
                ),
                description=collection_description,
            )

            if extra_metadata is not None:
                if extra_metadata.get("extra_metadata"):
                    ExtraMetadata(
                        logger_properties=self.logger_properties
                    ).collection(
                        collection=collection[collection_id],
                        extra_metadata=extra_metadata,
                    )
                else:
                    self.logger_properties["logger_level"] = "INFO"
                    self.logger_properties[
                        "logger_msg"
                    ] = "The `extra_metadata` is not activated. So, it does not add any extra metadata to the STAC collection."
                    # logger.Logger(self.logger_properties)

            catalog.add_child(collection[collection_id])

        # if collection_scientific is True:
        #     scientific_class = scientific.Scientific()
        #     scientific_class.collection(
        #         catalog.get_child(collection_id),
        #         collection_scientific,
        #     )
        return {
            "existed_items_id_list": existed_items_id_list,
            "collection": collection[collection_id],
            "collection_bbox_existed": collection["collection_bbox_existed"],
            "collection_interval_time_final_existed": collection[
                "collection_interval_time_final_existed"
            ],
        }

    def STACItem(
        self,
        url: str,
        catalog: pystac.Catalog,
        harvesting_vars: dict,
        Recognizer_output: Union[str, None],
        collection_id: str,
        aggregated_dataset_url: Union[str, None] = None,
        extension_properties: Union[dict, None] = None,
        asset_properties: Union[dict, None] = dict(),
        logger_properties: dict = dict(),
        extra_metadata: dict = dict(),
        stac_existence_collection: bool = False,
        collection_bbox_existed: Union[list, None] = None,
        collection_interval_time_final_existed: Union[list, None] = None,
    ):
        """
        This is a function for creating STAC item
        from harvested data in TDS dataset catalog.

        Args:
            url: The URL of the TDS catalog.
            catalog: The STAC catalog.
            harvesting_vars: The harvested data from TDS catalog.
            Recognizer_output: The output of the Recognizer class.
            collection_id: The ID of the STAC collection.
            aggregated_dataset_url: The URL of the aggregated dataset that
                whole of data is located there.
            extension_properties: The properties of the extensions.
            asset_properties: The properties of the assets.
            logger_properties: The properties of the logger. For more information please check the :class:`~tds2stac.logger.Logger` class.
        """
        # if "Seventh Scenario" in str(Recognizer_output):
        #     service_url_html = utils.xml2html(url)
        # else:
        #     service_url_html = (
        #         utils.xml2html(url)
        #         + "?dataset="
        #         + harvesting_vars["dataset"].get("ID")
        #     )

        if logger_properties is not None:
            self.logger_properties = logger_properties
        if (
            utils.replacement_func_collection_item_id(
                harvesting_vars["item_id"]
            )
            is None
            or harvesting_vars["item_footprint"] is None
            or harvesting_vars["item_bbox"] is None
            or harvesting_vars["modified_date_time"] is None
        ):
            self.logger_properties["logger_level"] = "CRITICAL"
            self.logger_properties[
                "logger_msg"
            ] = "One of `item_id`, `item_footprint`, `item_bbox` or `modified_date_time` is None, so you need to double check your dataset."
            logger.Logger(self.logger_properties)
            return
        item = pystac.Item(
            id=utils.replacement_func_collection_item_id(
                harvesting_vars["item_id"]
            ),
            geometry=geometry.mapping(harvesting_vars["item_footprint"]),
            bbox=harvesting_vars["item_bbox"],
            datetime=harvesting_vars["modified_date_time"],
            properties={},
        )

        if (
            extension_properties is not None
            and extension_properties["item_common_metadata"] is True
        ):
            common_metadata_class = common_metadata.CommonMetadata()
            common_metadata_class.item(
                item, harvesting_vars, self.logger_properties
            )

        #########################################
        # Adding web services as assets into items
        #########################################
        # profiler = cProfile.Profile()
        # profiler.enable()

        asset = assets.Assets()
        asset.item(
            harvesting_vars=harvesting_vars,
            item=item,
            Recognizer_output=Recognizer_output,
            url=url,
            aggregated_dataset_url=aggregated_dataset_url,
            asset_properties=asset_properties,
            logger_properties=self.logger_properties,
        )
        # profiler.disable()
        # stats = pstats.Stats(profiler).sort_stats('cumtime')
        # stats.print_stats()

        if extension_properties is not None:
            if extension_properties.get("item_extensions") is not None:
                for extension_property in extension_properties[
                    "item_extensions"
                ]:
                    # if (
                    #     isinstance(extension_property, str)
                    #     and extension_property == "item_scientific"
                    # ):
                    #     scientific_class = scientific.Scientific()
                    #     scientific_class.item(
                    #         item, extension_properties["item_scientific"]
                    #     )
                    if (
                        isinstance(extension_property, str)
                        and extension_property == "item_datacube_extension"
                    ):
                        datacube_class = datacube.Datacube()
                        datacube_class.item_extension(
                            item, harvesting_vars, self.logger_properties
                        )
                    if (
                        isinstance(extension_property, str)
                        and extension_property == "common_metadata"
                    ):
                        common_metadata_class = (
                            common_metadata.CommonMetadata()
                        )
                        common_metadata_class.item(
                            item, harvesting_vars, self.logger_properties
                        )
                    if isinstance(extension_property, tuple):
                        if (
                            len(extension_property) < 2
                            or len(extension_property) > 3
                        ):
                            self.logger_properties["logger_level"] = "ERROR"
                            self.logger_properties[
                                "logger_msg"
                            ] = "The length of `extension_property` tuple is less than 2 or greater than 3, so you need to double check your input"
                            logger.Logger(self.logger_properties)
                            return
                        elif len(extension_property) == 2:
                            caller_filename = inspect.stack()
                            script_path = caller_filename[2][1]
                            sys.path.append(script_path)
                            custom_module = os.path.splitext(
                                os.path.basename(caller_filename[2][1])
                            )[0]
                            imported_module = importlib.import_module(
                                custom_module
                            )

                        elif len(extension_property) == 3:
                            script_path_dir = os.path.dirname(
                                extension_property[2]
                            )
                            sys.path.append(script_path_dir)
                            custom_module = os.path.splitext(
                                os.path.basename(extension_property[2])
                            )[0]
                            # imported_module = importlib.import_module(
                            #     custom_module
                            # )
                            try:
                                spec = spec_from_loader(
                                    custom_module,
                                    SourceFileLoader(
                                        custom_module, extension_property[2]
                                    ),
                                )
                                imported_module = module_from_spec(spec)  # type: ignore
                                spec.loader.exec_module(imported_module)  # type: ignore
                            except Exception:
                                self.logger_properties[
                                    "logger_level"
                                ] = "ERROR"
                                self.logger_properties[
                                    "logger_msg"
                                ] = "Check your extension script path, third element of the tuple"
                                logger.Logger(self.logger_properties)
                                return

                        if "." in extension_property[1]:
                            extension_property_class_str = extension_property[
                                1
                            ].split(".")[0]
                            extension_property_function_str = (
                                extension_property[1].split(".")[1]
                            )
                            if hasattr(
                                imported_module, extension_property_class_str
                            ):
                                extension_property_class = getattr(
                                    imported_module,
                                    extension_property_class_str,
                                )
                                instance = extension_property_class()
                                if hasattr(
                                    instance, extension_property_function_str
                                ):
                                    a = getattr(
                                        instance,
                                        extension_property_function_str,
                                    )(item, harvesting_vars)

                                else:
                                    self.logger_properties[
                                        "logger_level"
                                    ] = "ERROR"
                                    self.logger_properties[
                                        "logger_msg"
                                    ] = "Check your extension function name, second element of the tuple"
                                    logger.Logger(self.logger_properties)
                            else:
                                self.logger_properties[
                                    "logger_level"
                                ] = "ERROR"
                                self.logger_properties[
                                    "logger_msg"
                                ] = "Check your extension class name, second element of the tuple"
                                logger.Logger(self.logger_properties)
                        else:
                            extension_property_function_str = (
                                extension_property[1]
                            )
                            if hasattr(
                                imported_module,
                                extension_property_function_str,
                            ):
                                a = getattr(
                                    imported_module,
                                    extension_property_function_str,
                                )(item, harvesting_vars)
                                print(a)

                            else:
                                self.logger_properties[
                                    "logger_level"
                                ] = "ERROR"
                                self.logger_properties[
                                    "logger_msg"
                                ] = "Check your extension function name, second element of the tuple"
                                logger.Logger(self.logger_properties)
                        sys.path.remove(script_path_dir)
        # if (
        #     extension_properties is not None
        #     and extension_properties["item_scientific"] is True
        # ):
        #     scientific_class = scientific.Scientific()
        #     scientific_class.item(
        #         item, extension_properties["item_scientific"]
        #     )
        if extra_metadata is not None:
            if extra_metadata.get("extra_metadata"):
                ExtraMetadata(logger_properties=self.logger_properties).item(
                    item=item,
                    extra_metadata=extra_metadata,
                    harvesting_vars=harvesting_vars,
                )
            else:
                self.logger_properties["logger_level"] = "INFO"
                self.logger_properties[
                    "logger_msg"
                ] = "The `extra_metadata` is not activated. So, it does not add any extra metadata to the STAC item."
                # logger.Logger(self.logger_properties)

        # applying datacube extension to items

        if (
            extension_properties is not None
            and extension_properties["item_datacube"] is True
        ):
            datacube_class = datacube.Datacube()
            datacube_class.item_extension(
                item, harvesting_vars, logger_properties=self.logger_properties
            )

        # Because Collection does not provide point coordination, this condition was applied.
        # TODO: Should be checked by collection_footprint_point None ir not SHOULD BE REFACTORED .....
        if (
            harvesting_vars["collection_bbox"][0]
            == harvesting_vars["collection_bbox"][2]
            or harvesting_vars["collection_bbox"][1]
            == harvesting_vars["collection_bbox"][3]
        ):
            harvesting_vars["collection_bbox"] = [
                harvesting_vars["collection_bbox"][0] - constants.epilon,
                harvesting_vars["collection_bbox"][1] - constants.epilon,
                harvesting_vars["collection_bbox"][2] + constants.epilon,
                harvesting_vars["collection_bbox"][3] + constants.epilon,
            ]

        spatial_extent = pystac.SpatialExtent(
            bboxes=[harvesting_vars["collection_bbox"]]
        )
        temporal_extent = pystac.TemporalExtent(
            intervals=[harvesting_vars["collection_interval_time_final"]]
        )
        # An empty condition for either Temporal or Spatial extent
        # TODO: To be refactored
        if (
            harvesting_vars["collection_bbox"] is None
            or harvesting_vars["collection_interval_time_final"] is None
        ):
            spatial_extent = pystac.SpatialExtent(bboxes=[0.0, 0.0])
            temporal_extent = pystac.TemporalExtent(
                intervals=[[datetime.utcnow(), datetime.utcnow()]]
            )

        if (
            stac_existence_collection is True
            and collection_bbox_existed is not None
            and collection_interval_time_final_existed is not None
        ):
            if (
                spatial_extent.bboxes[0][0] != 0.0
                and spatial_extent.bboxes[0][1] != 0.0
            ):
                collection_boundingbox = utils.merge_bboxes(
                    collection_bbox_existed, harvesting_vars["collection_bbox"]
                )
                spatial_extent = pystac.SpatialExtent(
                    bboxes=[collection_boundingbox]
                )
            else:
                spatial_extent = pystac.SpatialExtent(
                    bboxes=[collection_bbox_existed]
                )
            if (
                temporal_extent.intervals[0][0] != datetime.utcnow()
                and temporal_extent.intervals[0][1] != datetime.utcnow()
            ):
                collection_interval_time_final = utils.merge_intervals(
                    collection_interval_time_final_existed,
                    harvesting_vars["collection_interval_time_final"],
                )
                temporal_extent = pystac.TemporalExtent(
                    intervals=[collection_interval_time_final]
                )
            else:
                temporal_extent = pystac.TemporalExtent(
                    intervals=[collection_interval_time_final_existed]
                )

        # final_collection: pystac.Collection = catalog.get_child(collection_id)  # type: ignore
        # print(len(list(catalog.get_children())))
        # print(len(list(final_collection.get_children())))
        # print("Before adding item to collection")

        if catalog.get_child(collection_id) is not None:
            catalog.get_child(collection_id).extent = pystac.Extent(  # type: ignore
                spatial=spatial_extent,
                temporal=temporal_extent,
            )

            catalog.get_child(collection_id).add_item(item)  # type: ignore
            item = None  # type: ignore

        return {"item": item}

    def SaveCatalog(
        self, catalog, catalog_dir, logger_properties: dict = dict()
    ):
        self.logger_properties = logger_properties
        try:
            catalog.normalize_hrefs(os.path.join(catalog_dir, "stac"))
            catalog.save(catalog_type=pystac.CatalogType.SELF_CONTAINED)
            return True
        except Exception:
            ex_type, ex_value, ex_traceback = sys.exc_info()
            if ex_type is not None and ex_value is not None:
                self.logger_properties["logger_level"] = "CRITICAL"
                self.logger_properties["logger_msg"] = (
                    "The Collection doesn't contain bounding box and/or temporal information. Therefore the STAC-Catalog cannot be created. Review the input values. %s : %s"
                    % (
                        ex_type.__name__,
                        ex_value,
                    )
                )
                logger.Logger(self.logger_properties)
            else:
                self.logger_properties["logger_level"] = "CRITICAL"
                self.logger_properties[
                    "logger_msg"
                ] = "The Collection doesn't contain bounding box and/or temporal information. Therefore the STAC-Catalog cannot be created. Review the input values."
                logger.Logger(self.logger_properties)
            return False
            # print(traceback.format_exc())



# SPDX-FileCopyrightText: 2023 Karlsruher Institut für Technologie
#
# SPDX-License-Identifier: CC0-1.0

from typing import Union

from lxml import etree

from .. import logger, utils
from ..statics import constants
from . import recognizer


class NestedCollectionInspector(object):
    """
    This class will generate Collection IDs, Titles and
    their corresponding URLs for a presumed nested
    number originating from the Recognizer class
    in TDS. Only works for nested scenarios number 1,2,3,8
    and 9 in Recognizer class.
    The output will be a list of the tuples:
    (Root collection URL, Collection ID, Collection Title, corresponding subset URLs)

    Args:
        main_catalog_url (str): The URL of the TDS catalog
        nested_number (int, optional): Number of depth for nested datasets
        logger_properties (dict, optional): A dictionary for logger properties
        requests_properties (dict, optional): A dictionary for requests properties
    """

    main_catalog_url: str
    """
        The URL of the TDS catalog
    """
    nested_number: Union[int, None]
    """
        Number of depth for nested datasets
    """
    logger_properties: dict
    """
        A dictionary for logger properties. For more information see :class:`~tds2stac.logger.Logger`
    """
    requests_properties: dict
    """
        To obtain additional information on this topic, refer to
        the :class:`~tds2stac.TDS2STAC.requests_properties`. The default value is
        an empty dictionary.
    """

    def __init__(
        self,
        main_catalog_url: str,
        nested_number: Union[int, None] = None,
        logger_properties: dict = dict(),
        requests_properties: dict = dict(),
    ):
        self.requests_properties = requests_properties

        # using 'xml_processing' we get the XML contents of catalog URL.
        xml_url_catalog, id_catalog, xml = utils.xml_processing(
            main_catalog_url, self.requests_properties
        )
        if logger_properties is not None:
            self.logger_properties = logger_properties

        self.id_catalog = id_catalog
        self.xml_url_catalog = xml_url_catalog
        self.xml_url_catalog_temp = xml_url_catalog
        self.xml = xml
        self.all_urls: list = []
        self.corresponding_urls_ids: Union[list, None] = []
        self.depth_addresses: list = []

        self.nested_number = nested_number
        if self.nested_number is None:
            self.nested_number = 0
        if self.nested_number is not None:
            possible_nested_scenarios = [
                "First Scenario",
                "Third Scenario",
                "Second Scenario",
                "Eighth Scenario",
                "Ninth Scenario",
            ]

            if any(
                scenario
                in str(
                    recognizer.Recognizer(
                        main_catalog_url=self.xml_url_catalog,
                        requests_properties=self.requests_properties,
                    )
                )
                for scenario in possible_nested_scenarios
            ):
                if self.nested_number == 0:
                    # Getting the end point URLs of each dataset
                    self.final_collections_details_returner(
                        self.xml_url_catalog
                    )
                    # Finding the corresponding collection's URLs and IDs
                    collection_id = utils.replacement_func_collection_item_id(
                        xml_url_catalog
                    )
                    print(collection_id)
                    collection_title = utils.replacement_func(xml_url_catalog)
                    # Final variable than is a list of tuples and contains the
                    # corresponding collection's URLs and IDs and all related URLs
                    self.corresponding_urls_ids = [
                        (
                            xml_url_catalog,
                            collection_id,
                            collection_title,
                            self.all_urls,
                        )
                    ]
                else:
                    self.all_nested_dict: dict = {}
                    self.layer_dict: dict = {}
                    # Getting the list of all available datasets as a nested dictionary `self.all_nested_dict`
                    self.nested_dict_returner(
                        self.xml_url_catalog, self.all_nested_dict
                    )
                    # Getting the nested dictionary in a given depth `self.layer_dict`
                    self.layer_dict = self.n_level(
                        self.all_nested_dict, self.nested_number
                    )

                    # Getting the end point URLs of each dataset
                    # if isinstance(self.layer_dict, dict):
                    for i in self.layer_dict:
                        if isinstance(i, dict):
                            for k, v in i.items():
                                if isinstance(v, dict):
                                    self.end_point_url_extractor_dict(v)
                                if isinstance(v, list):
                                    self.end_point_url_extractor_list(v)
                                if isinstance(v, str):
                                    self.depth_addresses.append(v)
                        # rarely happens
                        if isinstance(i, list):
                            self.end_point_url_extractor_list(i)
                        if isinstance(i, str):
                            self.depth_addresses.append(i)

                    # Final variable than is a list of tuples and contains the
                    # corresponding collection's URLs and IDs and all related URLs
                    for i in self.depth_addresses:
                        self.all_urls = []
                        self.final_collections_details_returner(i)
                        collection_id = (
                            utils.replacement_func_collection_item_id(i)
                        )
                        collection_title = utils.replacement_func(i)
                        self.corresponding_urls_ids.append(
                            (i, collection_id, collection_title, self.all_urls)
                        )
                    if "Ninth Scenario" in str(
                        recognizer.Recognizer(
                            main_catalog_url=self.xml_url_catalog,
                            requests_properties=self.requests_properties,
                        )
                    ):
                        for i in self.corresponding_urls_ids:
                            for j in self.corresponding_urls_ids:
                                if i[0] in j[3]:
                                    j[3].remove(i[0])
            else:
                self.final_collections_details_returner(self.xml_url_catalog)
                # Finding the corresponding collection's URLs and IDs
                collection_id = utils.replacement_func_collection_item_id(
                    xml_url_catalog
                )
                print(collection_id)
                collection_title = utils.replacement_func(xml_url_catalog)
                # Final variable than is a list of tuples and contains the
                # corresponding collection's URLs and IDs and all related URLs
                self.corresponding_urls_ids = [
                    (
                        xml_url_catalog,
                        collection_id,
                        collection_title,
                        self.all_urls,
                    )
                ]
        self.logger_properties["logger_level"] = "INFO"
        self.logger_properties["logger_msg"] = self.corresponding_urls_ids
        logger.Logger(self.logger_properties)

    def end_point_url_extractor_dict(self, d: dict):
        """
        A function for extracting the end point URLs
        of a nested dictionary.

        Args:
            d (dict): A nested dictionary
        """
        for k, v in d.items():
            if isinstance(v, dict):
                # it should be v or d ?????
                self.end_point_url_extractor_dict(v)
            if isinstance(v, list):
                self.end_point_url_extractor_list(v)
            if isinstance(v, str):
                # if isinstance(self.depth_addresses, list):
                self.depth_addresses.append(v)

    def end_point_url_extractor_list(self, list_: list):
        """
        A function for extracting the end point URLs
        of a nested list.

        Args:
            list_ (list): A nested list
        """
        for i in list_:
            if isinstance(i, dict):
                self.end_point_url_extractor_dict(i)
            if isinstance(i, list):
                self.end_point_url_extractor_list(i)
            if isinstance(i, str):
                # if isinstance(self.depth_addresses, list):
                self.depth_addresses.append(i)

    def to_level(self, d: dict, layer: int):
        """
        A function for getting the a dictionary
        in a given depth. https://stackoverflow.com/a/68228562

        Args:
            d (dict): A nested dictionary
            layer (int): The depth of the dictionary

        """

        if layer:  # make sure level is not 0
            if not isinstance(d, (list, dict)):
                yield d  # not a dictionary or list, does not need to be traversed
            elif isinstance(d, dict):
                for (
                    a,
                    b,
                ) in (
                    d.items()
                ):  # element is a dictionary, have to transform it
                    if not (
                        n := list(self.to_level(b, layer - 1))
                    ):  # check this is the last level
                        yield a  # current level is `1`, so only yield back the key of the dictionary (no need to traverse the values, as we are out of depth)
                    else:  # at a level `l > 1`, so we need to transform the key's value
                        if isinstance(b, (str, list)) or (
                            len(b) == 1 and not isinstance(b, list)
                        ):
                            n = n[
                                0
                            ]  # original value `b` was a dictionary with a single key, a string, or a list
                        elif isinstance(b, dict) and all(
                            isinstance(i, dict) for i in n
                        ):
                            n = dict([j for k in n for j in k.items()])  # type: ignore  # `b` was a dictionary with more than one key, or a dictionary with multiple additional levels to traverse
                        yield {
                            a: n
                        }  # yield back the key and its transformed value as its own dictionary
            else:
                yield [
                    j for k in d for j in self.to_level(k, layer)
                ]  # value to transform is a list, simply map `to_level` to each element and yield the resulting list

    def n_level(self, d: dict, layer: int):
        """
        For decoding the generator object
        of `to_level` function. https://stackoverflow.com/a/68228562

        Args:
            d (dict): A nested dictionary
            layer (int): The depth of the dictionary
        """
        return list(self.to_level(d, layer))

    def nested_dict_returner(self, url: str, dict: dict):
        """
        A function for getting the nested dictionary
        of a given URL.

        Args:
            url (str): The URL of the TDS catalog
            dict (dict): A nested dictionary
        """

        xml_url_catalog, id_catalog, xml = utils.xml_processing(
            url, self.requests_properties
        )

        try:
            tree = etree.XML(xml)
        except BaseException:
            # self.logger.warning(
            #     "The Catalog is not reachable. Check the Catalg URL in the TDS"
            # )
            return (
                "The Catalog is not reachable. Check the Catalg URL in the TDS"
            )

        if tree.findall(".//{%s}catalogRef" % constants.unidata) != [] or (
            tree.findall(".//{%s}dataset[@urlPath]" % constants.unidata) != []
            and tree.findall(".//{%s}catalogRef" % constants.unidata) != []
        ):
            for child in tree.findall(
                ".//{%s}catalogRef" % constants.unidata
            ) + tree.findall(".//{%s}dataset[@urlPath]" % constants.unidata):
                if "catalogRef" in str(child):
                    # print(utils.references_urls(url, child.get("{%s}href" % constants.w3)))
                    dict[
                        utils.references_urls(
                            url, child.get("{%s}href" % constants.w3)
                        )
                    ] = {}
                    dic = dict[
                        utils.references_urls(
                            url, child.get("{%s}href" % constants.w3)
                        )
                    ]
                    self.nested_dict_returner(
                        utils.references_urls(
                            url, child.get("{%s}href" % constants.w3)
                        ),
                        dic,
                    )
                if (
                    "dataset" in str(child)
                    and len(
                        tree.findall(
                            ".//{%s}dataset[@urlPath]" % constants.unidata
                        )
                    )
                    == 1
                ):
                    dict[
                        url
                        + "?dataset="
                        + child.get("ID").replace("html", "xml")
                    ] = {}

                if (
                    "dataset" in str(child)
                    and len(
                        tree.findall(
                            ".//{%s}dataset[@urlPath]" % constants.unidata
                        )
                    )
                    > 1
                ):
                    dict[url.replace("html", "xml")] = {}

    def final_collections_details_returner(self, url: str):
        """
        A function for returning the URLs of input URL
        in First and Third cases in TDS

        Args:
            url (str): The URL of the TDS catalog
        """

        nested_scenarios = [
            "First Scenario",
            "Third Scenario",
            "Second Scenario",
            "Eighth Scenario",
            "Ninth Scenario",
        ]

        xml_url_catalog, id_catalog, xml = utils.xml_processing(
            url, self.requests_properties
        )

        try:
            tree = etree.XML(xml)
        except BaseException:
            # self.logger.warning(
            #     "The Catalog is not reachable. Check the Catalg URL in the TDS"
            # )
            return (
                "The Catalog is not reachable. Check the Catalg URL in the TDS"
            )

        if (
            tree.findall(".//{%s}dataset[@urlPath]" % constants.unidata) == []
            and tree.findall(".//{%s}catalogRef" % constants.unidata) != []
        ):
            for child in tree.findall(".//{%s}catalogRef" % constants.unidata):
                if "catalogRef" in str(child):
                    self.final_collections_details_returner(
                        utils.references_urls(
                            url, child.get("{%s}href" % constants.w3)
                        )
                    )
                    recog = recognizer.Recognizer(
                        main_catalog_url=utils.references_urls(
                            url, child.get("{%s}href" % constants.w3)
                        ),
                        requests_properties=self.requests_properties,
                    )
                    if utils.references_urls(
                        url, child.get("{%s}href" % constants.w3)
                    ) not in self.all_urls and not any(
                        scenario in str(recog) for scenario in nested_scenarios
                    ):
                        self.all_urls.append(
                            utils.references_urls(
                                url, child.get("{%s}href" % constants.w3)
                            )
                        )
        if (
            tree.findall(".//{%s}dataset[@urlPath]" % constants.unidata) != []
            and tree.findall(".//{%s}catalogRef" % constants.unidata) != []
        ):
            for child in tree.findall(
                ".//{%s}catalogRef" % constants.unidata
            ) + tree.findall(".//{%s}dataset[@urlPath]" % constants.unidata):
                if "catalogRef" in str(child):
                    self.final_collections_details_returner(
                        utils.references_urls(
                            url, child.get("{%s}href" % constants.w3)
                        )
                    )
                    recog = recognizer.Recognizer(
                        main_catalog_url=utils.references_urls(
                            url, child.get("{%s}href" % constants.w3)
                        ),
                        requests_properties=self.requests_properties,
                    )
                    if utils.references_urls(
                        url, child.get("{%s}href" % constants.w3)
                    ) not in self.all_urls and not any(
                        scenario in str(recog) for scenario in nested_scenarios
                    ):
                        self.all_urls.append(
                            utils.references_urls(
                                url, child.get("{%s}href" % constants.w3)
                            )
                        )
                if (
                    "dataset" in str(child)
                    and len(
                        tree.findall(
                            ".//{%s}dataset[@urlPath]" % constants.unidata
                        )
                    )
                    == 1
                ):
                    if (
                        url
                        + "?dataset="
                        + child.get("ID").replace("html", "xml")
                        not in self.all_urls
                    ):
                        self.all_urls.append(
                            url
                            + "?dataset="
                            + child.get("ID").replace("html", "xml")
                        )
                if (
                    "dataset" in str(child)
                    and len(
                        tree.findall(
                            ".//{%s}dataset[@urlPath]" % constants.unidata
                        )
                    )
                    > 1
                ):
                    if url.replace("html", "xml") not in self.all_urls:
                        self.all_urls.append(url.replace("html", "xml"))
        if (
            tree.findall(".//{%s}dataset[@urlPath]" % constants.unidata) != []
            and tree.findall(".//{%s}catalogRef" % constants.unidata) == []
        ):
            for child in tree.findall(".//{%s}catalogRef" % constants.unidata):
                self.all_urls = []

    def aslist(self):
        """
        A function for returning the list of tuples
        """
        return self.corresponding_urls_ids

    def __iter__(self):
        """
        A function for returning the iterator of tuples
        """
        return iter(self.aslist())



from datetime import datetime

# SPDX-FileCopyrightText: 2023 Karlsruher Institut für Technologie
#
# SPDX-License-Identifier: CC0-1.0
import pytz


class Temporal(object):
    def regulator(self, main_dict, temporal_format_by_dataname, data_name):
        errors = []
        warnings = []

        if temporal_format_by_dataname is not None:
            try:
                main_dict[
                    "temporal_extent_start_datetime"
                ] = datetime.strptime(
                    str(data_name), temporal_format_by_dataname
                ).strftime(
                    "%Y-%m-%dT%H:%M:%S.%fZ"
                )

            except Exception as e:
                start_datetime = self.safe_strip(
                    main_dict.get("temporal_extent_start_datetime")
                )
                if start_datetime:
                    main_dict[
                        "temporal_extent_start_datetime"
                    ] = start_datetime
                    warnings.append(
                        f"There is a problem with `temporal_format_by_dataname`. Revise your input. {e.__class__.__name__} : {e}"
                    )
                else:
                    errors.append(
                        f"Start DateTime is None or empty in the given dataset. Please review your `tag_config.json` file. {e.__class__.__name__} : {e}"
                    )

        else:
            start_datetime = self.safe_strip(
                main_dict.get("temporal_extent_start_datetime")
            )
            if start_datetime:
                main_dict["temporal_extent_start_datetime"] = start_datetime
            else:
                errors.append(
                    "Start DateTime is None or empty in the given dataset. Please review your `tag_config.json` file."
                )

        if temporal_format_by_dataname is not None:
            try:
                main_dict["temporal_extent_end_datetime"] = datetime.strptime(
                    str(data_name), temporal_format_by_dataname
                ).strftime("%Y-%m-%dT%H:%M:%S.%fZ")
            except Exception as e:
                end_datetime = self.safe_strip(
                    main_dict.get("temporal_extent_end_datetime")
                )
                if end_datetime:
                    main_dict["temporal_extent_end_datetime"] = end_datetime
                    warnings.append(
                        f"There is a problem with `temporal_format_by_dataname`. Revise your input. {e.__class__.__name__} : {e}"
                    )
                else:
                    errors.append(
                        f"End DateTime is None or empty in the given dataset. Please review your `tag_config.json` file. {e.__class__.__name__} : {e}"
                    )

        else:
            end_datetime = self.safe_strip(
                main_dict.get("temporal_extent_end_datetime")
            )
            if end_datetime:
                main_dict["temporal_extent_end_datetime"] = end_datetime
            else:
                errors.append(
                    "End DateTime is None or empty in the given dataset. Please review your `tag_config.json` file."
                )

        # Check if both start and end date-times are present and not just whitespace
        if all(
            main_dict.get(key) and not main_dict.get(key).isspace()
            for key in [
                "temporal_extent_start_datetime",
                "temporal_extent_end_datetime",
            ]
        ):
            # Parse and append start and end date-times to collection_interval_time
            main_dict["collection_interval_time"].extend(
                [
                    self.parse_datetime_with_fallback(
                        main_dict["temporal_extent_start_datetime"],
                        "%Y-%m-%dT%H:%M:%S.%fZ",
                        "%Y-%m-%dT%H:%M:%SZ",
                        pytz.utc,
                    ),
                    self.parse_datetime_with_fallback(
                        main_dict["temporal_extent_end_datetime"],
                        "%Y-%m-%dT%H:%M:%S.%fZ",
                        "%Y-%m-%dT%H:%M:%SZ",
                        pytz.utc,
                    ),
                ]
            )
            # Sort the collection_interval_time list
            main_dict["collection_interval_time"].sort()
        else:
            # Append error if either start or end datetime is missing or only whitespace
            errors.append(
                "Start or End DateTime is None or empty in the given dataset. Please review your `tag_config.json` file."
            )

        if errors != [] and warnings != []:
            return errors, warnings
        elif errors != [] and warnings == []:
            return errors, None
        elif errors == [] and warnings != []:
            return None, warnings
        else:
            return None, None

    # Function to parse datetime with a fallback format
    def parse_datetime_with_fallback(
        self, datetime_str, primary_format, fallback_format, tzinfo
    ):
        try:
            return datetime.strptime(datetime_str, primary_format).replace(
                tzinfo=tzinfo
            )
        except ValueError:
            return datetime.strptime(datetime_str, fallback_format).replace(
                tzinfo=tzinfo
            )

    def safe_strip(self, value):
        return value.strip() if value is not None else ""



# SPDX-FileCopyrightText: 2023 Karlsruher Institut für Technologie
#
# SPDX-License-Identifier: EUPL-1.2
# isort: skip_file

"""TDS2STAC

TDS2STAC
"""

__all__ = [
    "__version__",
    "TDS2STACIntegrator",
    "Logger",
    "JSONFileWebServiceListScraper",
    "WebServiceListScraper",
    "WebServiceContentScraper",
    "ExistenceValidator",
    "CollectionHarvester",
    "ItemHarvester",
    "STACCreator",
    "NestedCollectionInspector",
    "Recognizer",
    "Spatial",
    "Temporal",
    "Datacube",
    "Verifier",
    "Thumbnails",
]

from . import _version

from tds2stac.tds2stac import TDS2STACIntegrator

from tds2stac.analysers.existence_validator import ExistenceValidator
from tds2stac.logger import Logger

from tds2stac.webservices.core import (
    JSONFileWebServiceListScraper,
    WebServiceContentScraper,
    WebServiceListScraper,
)

from tds2stac.harvester import CollectionHarvester

from tds2stac.harvester import ItemHarvester

from tds2stac.creator import STACCreator

from tds2stac.analysers.nested_collections import NestedCollectionInspector
from tds2stac.analysers.recognizer import Recognizer
from tds2stac.dimensions.spatial import Spatial
from tds2stac.dimensions.temporal import Temporal

from tds2stac.extensions.datacube import Datacube
from tds2stac.analysers.properties_verifier import Verifier
from tds2stac.thumbnails import Thumbnails


__version__ = _version.get_versions()["version"]

__author__ = "Mostafa Hadizadeh"
__copyright__ = "2023 Karlsruher Institut für Technologie"
__credits__ = [
    "Mostafa Hadizadeh",
]
__license__ = "EUPL-1.2"

__maintainer__ = "Mostafa Hadizadeh"
__email__ = "mostafa.hadizadeh@kit.edu"

__status__ = "Pre-Alpha"

from . import _version

__version__ = _version.get_versions()["version"]



# SPDX-FileCopyrightText: 2023 Karlsruher Institut für Technologie
#
# SPDX-License-Identifier: EUPL-1.2

"""Test file for imports."""


def test_package_import():
    """Test the import of the main package."""
    import tds2stac  # noqa: F401
    from tds2stac.analysers.existence_validator import (  # noqa: F401
        ExistenceValidator,
    )
    from tds2stac.analysers.nested_collections import (  # noqa: F401
        NestedCollectionInspector,
    )
    from tds2stac.analysers.properties_verifier import Verifier  # noqa: F401
    from tds2stac.analysers.recognizer import Recognizer  # noqa: F401
    from tds2stac.creator import STACCreator  # noqa: F401
    from tds2stac.dimensions.spatial import Spatial  # noqa: F401
    from tds2stac.dimensions.temporal import Temporal  # noqa: F401
    from tds2stac.extensions.datacube import Datacube  # noqa: F401
    from tds2stac.harvester import CollectionHarvester  # noqa: F401
    from tds2stac.harvester import ItemHarvester  # noqa: F401
    from tds2stac.logger import Logger  # noqa: F401
    from tds2stac.tds2stac import TDS2STACIntegrator  # noqa: F401
    from tds2stac.thumbnails import Thumbnails  # noqa: F401
    from tds2stac.webservices.core import (  # noqa: F401
        JSONFileWebServiceListScraper,
        WebServiceContentScraper,
        WebServiceListScraper,
    )



# SPDX-FileCopyrightText: 2023 Karlsruher Institut für Technologie
#
# SPDX-License-Identifier: CC0-1.0

import json
import random

import pystac
import requests
from lxml import etree

from . import logger, utils
from .statics import constants


class Thumbnails(object):
    """
    This class is used to create thumbnail images for STAC-Collections and STAC-Items.
    """

    def collection(
        self,
        collection_thumbnail: str,
        collection_overview: str,
        services: etree.Element,
        dataset: dict,
        harvesting_vars: dict,
        collection_id: str,
        url: str,
        catalog: pystac.Catalog,
        collection_thumbnail_link: str,
        collection_overview_link: str,
        logger_properties: dict = dict(),
    ):
        """
        A function to create thumbnail images for STAC-Collections.

        Args:
            collection_thumbnail (str): The type of thumbnail image for STAC-Collections. It can be `wms` or `link`.
            collection_overview (str): The type of overview image for STAC-Collections. It can be `wms` or `link`.
            services (list): A list of services for STAC-Collections.
            dataset (dict): A dictionary of dataset information.
            harvesting_vars (dict): A dictionary of harvesting variables.
            collection_id (str): The ID of STAC-Collections.
            url (str): The URL of STAC-Catalog.
            catalog (pystac.Catalog): A STAC-Catalog.
            collection_thumbnail_link (str): The link of thumbnail image for STAC-Collections when `collection_thumbnail` or `collection_overview` set as `link`.
            collection_overview_link (str): The link of overview image for STAC-Collections when `collection_thumbnail` or `collection_overview` set as `link`.
            logger_properties (dict): A dictionary of logger properties. For more information, please see :class:`~tds2stac.logger.Logger` class.

        """
        collection: pystac.Collection = constants.empty_collection
        if logger_properties is not None:
            self.logger_properties = logger_properties

        if collection_thumbnail == "wms" or collection_overview == "wms":
            rand_idx = random.randrange(len(constants.colorbars))
            random_num = constants.colorbars[rand_idx]
            for service in services:
                if (
                    service.get("serviceType") == "Compound"
                    or service.get("serviceType") == "compound"
                ):
                    for s in service.findall(
                        "{%s}service" % constants.unidata
                    ):
                        if s.get("name") == "wms":
                            service_url = utils.references_urls(
                                url, s.get("base")
                            ) + dataset.get("urlPath")
                            minmax_url = (
                                service_url
                                + "?VERSION=1.1.1&REQUEST=GetMetadata&item=minmax&SRS=EPSG%3A4326&LAYERS="
                                + harvesting_vars["variable_ids"][0]
                                + "&BBOX="
                                + harvesting_vars["horizontal_extent_lon_min"]
                                + ","
                                + str(
                                    float(
                                        harvesting_vars[
                                            "horizontal_extent_lat_min"
                                        ]
                                    )
                                )
                                + ","
                                + harvesting_vars["horizontal_extent_lon_max"]
                                + ","
                                + str(
                                    float(
                                        harvesting_vars[
                                            "horizontal_extent_lat_max"
                                        ]
                                    )
                                )
                                + "&WIDTH=1000&HEIGHT=1000&FORMAT=image/png"
                                + "&TIME="
                                + harvesting_vars[
                                    "temporal_extent_start_datetime"
                                ]
                            )
                            get_minmax_head = requests.head(
                                minmax_url, timeout=10
                            )
                            if get_minmax_head.status_code == 200:
                                get_minmax = requests.get(
                                    minmax_url,
                                    None,
                                    verify=False,
                                    timeout=10,
                                )
                                minmax = json.loads(get_minmax.text)

                                min_color = minmax["min"]
                                max_color = minmax["max"]

                                thumbnail_url = (
                                    service_url
                                    + "?SERVICE=WMS&VERSION=1.1.1&REQUEST=GetMap&SRS=EPSG%3A4326&TRANSPARENT=TRUE&ABOVEMAXCOLOR=extend&BELOWMINCOLOR=extend&BGCOLOR=extend&LOGSCALE=false&&COLORSCALERANGE="
                                    + str(min_color)
                                    + ","
                                    + str(max_color)
                                    + "&STYLES=default-scalar/"
                                    + random_num
                                    + "&LAYERS="
                                    + harvesting_vars["variable_ids"][0]
                                    + "&BBOX="
                                    + harvesting_vars[
                                        "horizontal_extent_lon_min"
                                    ]
                                    + ","
                                    + str(
                                        float(
                                            harvesting_vars[
                                                "horizontal_extent_lat_min"
                                            ]
                                        )
                                    )
                                    + ","
                                    + harvesting_vars[
                                        "horizontal_extent_lon_max"
                                    ]
                                    + ","
                                    + str(
                                        float(
                                            harvesting_vars[
                                                "horizontal_extent_lat_max"
                                            ]
                                        )
                                    )
                                    + "&WIDTH=1000&HEIGHT=1000&FORMAT=image/png"
                                    + "&TIME="
                                    + harvesting_vars[
                                        "temporal_extent_start_datetime"
                                    ]
                                )
                            else:
                                self.logger_properties[
                                    "logger_level"
                                ] = "WARNING"
                                self.logger_properties["logger_msg"] = (
                                    "The getminmax thumbnail argument for "
                                    + collection_id
                                    + " is not available. It will be replaced by the default thumbnail."
                                )
                                logger.Logger(self.logger_properties)
                                thumbnail_url = (
                                    service_url
                                    + "?SERVICE=WMS&VERSION=1.1.1&REQUEST=GetMap&SRS=EPSG%3A4326&TRANSPARENT=TRUE&ABOVEMAXCOLOR=extend&BELOWMINCOLOR=extend&BGCOLOR=extend&LOGSCALE=false&STYLES=default-scalar/"
                                    + random_num
                                    + "&LAYERS="
                                    + harvesting_vars["variable_ids"][0]
                                    + "&BBOX="
                                    + harvesting_vars[
                                        "horizontal_extent_lon_min"
                                    ]
                                    + ","
                                    + str(
                                        float(
                                            harvesting_vars[
                                                "horizontal_extent_lat_min"
                                            ]
                                        )
                                    )
                                    + ","
                                    + harvesting_vars[
                                        "horizontal_extent_lon_max"
                                    ]
                                    + ","
                                    + str(
                                        float(
                                            harvesting_vars[
                                                "horizontal_extent_lat_max"
                                            ]
                                        )
                                    )
                                    + "&WIDTH=1000&HEIGHT=1000&FORMAT=image/png"
                                    + "&TIME="
                                    + harvesting_vars[
                                        "temporal_extent_start_datetime"
                                    ]
                                )
                            head_of_thumbnail = requests.head(
                                thumbnail_url, timeout=10
                            )
                            if head_of_thumbnail.status_code == 200:
                                if collection_thumbnail in ["wms"]:
                                    collection = catalog.get_child(collection_id)  # type: ignore
                                    collection.add_asset(
                                        key="thumbnail",
                                        asset=pystac.Asset(
                                            href=thumbnail_url,
                                            roles=["thumbnail"],
                                            # title=without_slash,
                                            media_type=pystac.MediaType.PNG,
                                        ),
                                    )
                                if collection_overview in ["wms"]:
                                    collection = catalog.get_child(collection_id)  # type: ignore
                                    collection.add_asset(
                                        key="rendered_preview",
                                        asset=pystac.Asset(
                                            href=thumbnail_url,
                                            roles=["visual"],
                                            # title=collection_id,
                                            media_type=pystac.MediaType.PNG,
                                        ),
                                    )
                            else:
                                self.logger_properties[
                                    "logger_level"
                                ] = "ERROR"
                                self.logger_properties["logger_msg"] = (
                                    "The thumbnail image for "
                                    + collection_id
                                    + " is not available."
                                )
                                logger.Logger(self.logger_properties)

        if (
            collection_thumbnail == "link"
            and collection_thumbnail_link is not None
        ):
            collection = catalog.get_child(collection_id)  # type: ignore
            collection.add_asset(
                key="thumbnail",
                asset=pystac.Asset(
                    href=collection_thumbnail_link,
                    roles=["thumbnail"],
                    title=collection_id,
                    media_type=pystac.MediaType.PNG,
                ),
            )
        if (
            collection_overview == "link"
            and collection_overview_link is not None
        ):
            collection = catalog.get_child(collection_id)  # type: ignore
            collection.add_asset(
                key="rendered_preview",
                asset=pystac.Asset(
                    href=collection_overview_link,
                    roles=["visual"],
                    # title=collection_id,
                    media_type=pystac.MediaType.PNG,
                ),
            )

    def item(
        self,
        service: etree.Element,
        dataset: dict,
        harvesting_vars: dict,
        url: str,
        item: pystac.Item,
        item_thumbnail: bool,
        item_overview: bool,
        item_getminmax_thumbnail: bool,
        logger_properties: dict = dict(),
    ):
        """
        A function to create thumbnail images for STAC-Items.

        Args:
            service (list): A list of services for STAC-Items.
            dataset (dict): A dictionary of dataset information.
            harvesting_vars (dict): A dictionary of harvesting variables.
            url (str): The URL of STAC-Catalog.
            item (pystac.Item): A STAC-Item.
            item_thumbnail (bool): A boolean to create thumbnail image for STAC-Items.
            item_overview (bool): A boolean to create overview image for STAC-Items.
            item_getminmax_thumbnail (bool): A boolean to create thumbnail image for STAC-Items based on minmax.
            logger_properties (dict): A dictionary of logger properties. For more information, please see :class:`~tds2stac.logger.Logger` class.
        """
        rand_idx = random.randrange(len(constants.colorbars))
        random_num = constants.colorbars[rand_idx]
        if logger_properties is not None:
            self.logger_properties = logger_properties
        for s in service.findall("{%s}service" % constants.unidata):
            if (
                s.get("name") == "wms"
                and (item_thumbnail is True or item_overview is True)
                and item_getminmax_thumbnail is not True
            ):
                service_url = utils.references_urls(
                    url, s.get("base")
                ) + dataset.get("urlPath")
                thumbnail_url = (
                    service_url
                    + "?SERVICE=WMS&VERSION=1.1.1&REQUEST=GetMap&SRS=EPSG%3A4326&TRANSPARENT=TRUE&ABOVEMAXCOLOR=extend&BELOWMINCOLOR=extend&BGCOLOR=extend&LOGSCALE=false&STYLES=default-scalar/"
                    + random_num
                    + "&LAYERS="
                    + harvesting_vars["variable_ids"][0]
                    + "&BBOX="
                    + harvesting_vars["horizontal_extent_lon_min"]
                    + ","
                    + str(float(harvesting_vars["horizontal_extent_lat_min"]))
                    + ","
                    + harvesting_vars["horizontal_extent_lon_max"]
                    + ","
                    + str(float(harvesting_vars["horizontal_extent_lat_max"]))
                    + "&WIDTH=1000&HEIGHT=1000&FORMAT=image/png"
                    + "&TIME="
                    + harvesting_vars["temporal_extent_start_datetime"]
                )
                item_request = requests.head(thumbnail_url, timeout=10)
                if item_request.status_code == 200:
                    if item_thumbnail is True:
                        item.add_asset(
                            key="thumbnail",
                            asset=pystac.Asset(
                                href=thumbnail_url,
                                roles=["thumbnail"],
                                # title=without_slash,
                                media_type=pystac.MediaType.PNG,
                            ),
                        )
                    if item_overview is True:
                        item.add_asset(
                            key="rendered_preview",
                            asset=pystac.Asset(
                                href=thumbnail_url,
                                roles=["overview"],
                                # title=without_slash,
                                media_type=pystac.MediaType.PNG,
                            ),
                        )

                else:
                    self.logger_properties["logger_level"] = "ERROR"
                    self.logger_properties["logger_msg"] = (
                        "The thumbnail image for "
                        + item.id
                        + " is not available."
                    )
                    logger.Logger(self.logger_properties)

            if (
                s.get("name") == "wms"
                and (item_thumbnail is True or item_overview is True)
                and item_getminmax_thumbnail is True
            ):
                service_url = utils.references_urls(
                    url, s.get("base")
                ) + dataset.get("urlPath")
                minmax_url = (
                    service_url
                    + "?VERSION=1.1.1&REQUEST=GetMetadata&item=minmax&SRS=EPSG%3A4326&LAYERS="
                    + harvesting_vars["variable_ids"][0]
                    + "&BBOX="
                    + harvesting_vars["horizontal_extent_lon_min"]
                    + ","
                    + str(float(harvesting_vars["horizontal_extent_lat_min"]))
                    + ","
                    + harvesting_vars["horizontal_extent_lon_max"]
                    + ","
                    + str(float(harvesting_vars["horizontal_extent_lat_max"]))
                    + "&WIDTH=1000&HEIGHT=1000&FORMAT=image/png"
                    + "&TIME="
                    + harvesting_vars["temporal_extent_start_datetime"]
                )

                get_minmax_head = requests.head(minmax_url, timeout=10)
                if get_minmax_head.status_code == 200:
                    get_minmax = requests.get(minmax_url, None, verify=False)

                    minmax = json.loads(get_minmax.text)

                    min_color = minmax["min"]
                    max_color = minmax["max"]

                    thumbnail_url = (
                        service_url
                        + "?SERVICE=WMS&VERSION=1.1.1&REQUEST=GetMap&SRS=EPSG%3A4326&TRANSPARENT=TRUE&ABOVEMAXCOLOR=extend&BELOWMINCOLOR=extend&BGCOLOR=extend&LOGSCALE=false&&COLORSCALERANGE="
                        + str(min_color)
                        + ","
                        + str(max_color)
                        + "&STYLES=default-scalar/"
                        + random_num
                        + "&LAYERS="
                        + harvesting_vars["variable_ids"][0]
                        + "&BBOX="
                        + harvesting_vars["horizontal_extent_lon_min"]
                        + ","
                        + str(
                            float(harvesting_vars["horizontal_extent_lat_min"])
                        )
                        + ","
                        + harvesting_vars["horizontal_extent_lon_max"]
                        + ","
                        + str(
                            float(harvesting_vars["horizontal_extent_lat_max"])
                        )
                        + "&WIDTH=1000&HEIGHT=1000&FORMAT=image/png"
                        + "&TIME="
                        + harvesting_vars["temporal_extent_start_datetime"]
                    )
                else:
                    self.logger_properties["logger_level"] = "WARNING"
                    self.logger_properties["logger_msg"] = (
                        "The getminmax thumbnail argument for "
                        + item.id
                        + " is not available. It will be replaced by the default thumbnail."
                    )
                    logger.Logger(self.logger_properties)

                    thumbnail_url = (
                        service_url
                        + "?SERVICE=WMS&VERSION=1.1.1&REQUEST=GetMap&SRS=EPSG%3A4326&TRANSPARENT=TRUE&ABOVEMAXCOLOR=extend&BELOWMINCOLOR=extend&BGCOLOR=extend&LOGSCALE=false&STYLES=default-scalar/"
                        + random_num
                        + "&LAYERS="
                        + harvesting_vars["variable_ids"][0]
                        + "&BBOX="
                        + harvesting_vars["horizontal_extent_lon_min"]
                        + ","
                        + str(
                            float(harvesting_vars["horizontal_extent_lat_min"])
                        )
                        + ","
                        + harvesting_vars["horizontal_extent_lon_max"]
                        + ","
                        + str(
                            float(harvesting_vars["horizontal_extent_lat_max"])
                        )
                        + "&WIDTH=1000&HEIGHT=1000&FORMAT=image/png"
                        + "&TIME="
                        + harvesting_vars["temporal_extent_start_datetime"]
                    )
                item_request = requests.head(thumbnail_url, timeout=10)
                if item_request.status_code == 200:
                    if item_thumbnail is True:
                        item.add_asset(
                            key="thumbnail",
                            asset=pystac.Asset(
                                href=thumbnail_url,
                                roles=["thumbnail"],
                                # title=without_slash,
                                media_type=pystac.MediaType.PNG,
                            ),
                        )
                    if item_overview is True:
                        item.add_asset(
                            key="rendered_preview",
                            asset=pystac.Asset(
                                href=thumbnail_url,
                                roles=["overview"],
                                # title=without_slash,
                                media_type=pystac.MediaType.PNG,
                            ),
                        )
                else:
                    self.logger_properties["logger_level"] = "ERROR"
                    self.logger_properties["logger_msg"] = (
                        "The thumbnail image for "
                        + item.id
                        + " is not available."
                    )
                    logger.Logger(self.logger_properties)



# SPDX-FileCopyrightText: 2023 Karlsruher Institut für Technologie
#
# SPDX-License-Identifier: CC0-1.0


from typing import Union
from urllib import parse as urlparse
from urllib.parse import quote_plus

import pystac
import requests

from . import utils
from .statics import constants
from .thumbnails import Thumbnails


class Assets:
    """
    This class is tasked with the responsibility of incorporating
    assets into STAC-Collections and STAC-Items.
    """

    def item(
        self,
        harvesting_vars: dict,
        item: pystac.Item,
        url: str,
        Recognizer_output: Union[str, None] = None,
        aggregated_dataset_url: Union[str, None] = None,
        asset_properties: Union[dict, None] = dict(),
        logger_properties: dict = dict(),
    ):
        """
        This is a function for adding assets to STAC-Items.

        Args:
            harvesting_vars (dict): A dictionary containing the variables required for harvesting.
            item (pystac.Item): A STAC-Item.
            Recognizer_output (dict): A dictionary containing the scenario output of the :class:`~tds2stac.Recognizer`.
            aggregated_dataset_url (str): The URL of the aggregated dataset.
            url (str): The URL of the catalog. It will be used for thumbnails
            asset_properties (dict): A dictionary containing the properties of the assets for more information refer to :class:`~tds2stac.TDS2STACIntegrator.asset_properties`.
            logger_properties (dict): A dictionary containing the properties of the logger for more information refer to :class:`~tds2stac.TDS2STACIntegrator.logger_properties`.
        """
        # Adding web services as assets into items
        if logger_properties is not None:
            self.logger_properties = logger_properties
        media_type_: Union[str, pystac.MediaType] = ""
        # for service in harvesting_vars["services"]:
        #     print(harvesting_vars["services_tuples"])

        #     if (
        #         service.get("serviceType") == "Compound"
        #         or service.get("serviceType") == "compound"
        #     ):
        if (
            asset_properties is not None
            and asset_properties.get("jupyter_notebook") is True
        ):
            catalog_path = urlparse.urlsplit(url).path.replace(
                "/thredds/catalog/", ""
            )
            additional = "/thredds/notebook/"
            service_url_notebook = utils.references_urls(
                url, additional
            ) + harvesting_vars["dataset"].get("ID")
            item.add_asset(
                key="jupyter_notebook",
                asset=pystac.Asset(
                    href=service_url_notebook
                    + "?catalog=%s&filename=%s"
                    % (catalog_path, "default_viewer.ipynb"),
                    title="Jupyter Notebook",
                    media_type=pystac.MediaType.HTML,
                ),
            )
        for s in harvesting_vars["services_tuples"]:
            service_url = utils.references_urls(url, s[0]) + harvesting_vars[
                "dataset"
            ].get("urlPath")

            if s[1] is not None:
                service_url += s[1]
            name_str: str = str(s[2])
            if s[2] in [
                "iso",
                "ncml",
                "uddc",
            ]:
                service_url += "?dataset=%s&&catalog=%s" % (
                    harvesting_vars["catalog_id"],
                    quote_plus(harvesting_vars["catalog_url"]),
                )
            elif s[2] in [
                "wms",
            ]:
                if (
                    asset_properties is not None
                    and asset_properties.get("explore_data") is True
                    and asset_properties.get("verify_explore_data") is not True
                ):
                    item.add_asset(
                        key="explore_data",
                        asset=pystac.Asset(
                            href=utils.references_urls(
                                url, "/thredds/Godiva.html"
                            )
                            + "?server="
                            + service_url,
                            title="Explore Data",
                            media_type=pystac.MediaType.HTML,
                        ),
                    )
                elif (
                    asset_properties is not None
                    and asset_properties.get("explore_data") is True
                    and asset_properties.get("verify_explore_data") is True
                ):
                    head = requests.head(
                        service_url + "?request=GetMetadata&item=menu"
                    )
                    if head.status_code == 200:
                        item.add_asset(
                            key="explore_data",
                            asset=pystac.Asset(
                                href=utils.references_urls(
                                    url, "/thredds/Godiva.html"
                                )
                                + "?server="
                                + service_url,
                                title="Explore Data",
                                media_type=pystac.MediaType.HTML,
                            ),
                        )
                service_url += (
                    "?service=WMS&version=1.3.0&request=GetCapabilities"
                )
            elif (
                s[2]
                in [
                    "http",
                ]
                and "Seventh Scenario" in str(Recognizer_output)
                and aggregated_dataset_url is not None
            ):
                service_url = aggregated_dataset_url
                media_type_ = pystac.MediaType.HTML
            elif (
                s[2]
                in [
                    "http",
                ]
                and "Seventh Scenario" in str(Recognizer_output)
                and aggregated_dataset_url is None
            ):
                # service_url += "?service=WMS&version=1.3.0&request=GetCapabilities"
                media_type_ = pystac.MediaType.HTML

            if s[2] in ["odap"]:
                service_url += ".html"
            # Determinatio of Media Type
            if s[2] in [
                "iso",
                "ncml",
                "wms",
                "wcs",
                "wfs",
                "sos",
            ]:
                media_type_ = pystac.MediaType.XML
            elif s[2] in ["http"] and "Seventh Scenario" not in str(
                Recognizer_output
            ):
                media_type_ = "application/netcdf"
            elif s[2] in [
                "dap4",
                "odap",
                "uddc",
            ]:
                media_type_ = pystac.MediaType.HTML
            else:
                media_type_ = pystac.MediaType.TEXT

            if asset_properties is None or (
                (
                    asset_properties.get("assets_list_allowed") is None
                    or asset_properties.get("assets_list_allowed") == []
                    or isinstance(
                        asset_properties.get("assets_list_allowed"),
                        str,
                    )
                )
                and (
                    asset_properties.get("assets_list_avoided") is None
                    or asset_properties.get("assets_list_avoided") == []
                    or isinstance(
                        asset_properties.get("assets_list_avoided"),
                        str,
                    )
                )
            ):
                item.add_asset(
                    key=s[2],
                    asset=pystac.Asset(
                        href=service_url,
                        # title=without_slash,
                        media_type=media_type_,
                    ),
                )

            elif (
                asset_properties is not None
                and (
                    asset_properties.get("assets_list_allowed") is not None
                    and isinstance(
                        asset_properties.get("assets_list_allowed"),
                        list,
                    )
                )
                and (
                    asset_properties.get("assets_list_avoided") is not None
                    and isinstance(
                        asset_properties.get("assets_list_avoided"),
                        list,
                    )
                )
            ):
                if name_str in (  # type: ignore
                    asset_properties.get("assets_list_allowed")
                ) and name_str not in (  # type: ignore
                    asset_properties.get("assets_list_avoided")
                ):
                    item.add_asset(
                        key=s[2],
                        asset=pystac.Asset(
                            href=service_url,
                            # title=without_slash,
                            media_type=media_type_,
                        ),
                    )
            elif (
                asset_properties is not None
                and (
                    asset_properties.get("assets_list_allowed") is None
                    or asset_properties.get("assets_list_allowed") == []
                    or isinstance(
                        asset_properties.get("assets_list_allowed"),
                        str,
                    )
                )
                and (
                    asset_properties.get("assets_list_avoided") is not None
                    and isinstance(
                        asset_properties.get("assets_list_avoided"),
                        list,
                    )
                )
            ):
                if name_str not in (asset_properties.get("assets_list_avoided")):  # type: ignore
                    item.add_asset(
                        key=s[2],
                        asset=pystac.Asset(
                            href=service_url,
                            # title=without_slash,
                            media_type=media_type_,
                        ),
                    )
            elif (
                asset_properties is not None
                and (
                    asset_properties.get("assets_list_allowed") is not None
                    and isinstance(
                        asset_properties.get("assets_list_allowed"),
                        list,
                    )
                )
                and (
                    asset_properties.get("assets_list_avoided") is None
                    or asset_properties.get("assets_list_avoided") == []
                    or isinstance(
                        asset_properties.get("assets_list_avoided"),
                        str,
                    )
                )
            ):
                if name_str in (asset_properties.get("assets_list_allowed")):  # type: ignore
                    item.add_asset(
                        key=s[2],
                        asset=pystac.Asset(
                            href=service_url,
                            # title=without_slash,
                            media_type=media_type_,
                        ),
                    )
            if asset_properties is not None:
                if asset_properties.get("item_thumbnail") is not None:
                    thumbnail = Thumbnails()
                    thumbnail.item(
                        harvesting_vars["services"][0],
                        harvesting_vars["dataset"],
                        harvesting_vars,
                        url,
                        item,
                        asset_properties["item_thumbnail"],
                        asset_properties["item_overview"],
                        asset_properties["item_getminmax_thumbnail"],
                        logger_properties=self.logger_properties,
                    )
                if asset_properties.get("item_custom_asset") is not None:
                    for asset in asset_properties["item_custom_asset"]:
                        item.add_asset(
                            key=asset.get("key"),
                            asset=pystac.Asset(
                                href=asset.get("href"),
                                title=asset.get("title")
                                if asset.get("title") is not None
                                else asset.get("key"),
                                roles=asset["roles"]
                                if asset.get("roles") is not None
                                else ["data"],
                                media_type=asset["media_type"]
                                if asset.get("media_type") is not None
                                else pystac.MediaType.TEXT,
                            ),
                        )
            # print(s.get("base"), s.get("suffix"), s.get("name"),service_url)
            # s.clear()
            # service.clear()

    def collection(
        self,
        harvesting_vars: dict,
        collection_dict: dict,
        stac_catalog: pystac.Catalog,
        asset_properties: dict = dict(),
        logger_properties: dict = dict(),
    ):
        """
        This is a function for adding assets to STAC-Collections.

        Args:
            asset_properties (dict): A dictionary containing the properties of the assets for more information refer to :class:`~tds2stac.TDS2STACIntegrator.asset_properties`.
            harvesting_vars (dict): A dictionary containing the variables required for harvesting.
            collection_dict (dict): A dictionary containing the properties of the collection.
            stac_catalog (pystac.Catalog): A STAC-Catalog.
            logger_properties (dict): A dictionary containing the properties of the logger for more information refer to :class:`~tds2stac.TDS2STACIntegrator.logger_properties`.


        """
        if logger_properties is not None:
            self.logger_properties = logger_properties

        if asset_properties is not None:
            if (
                asset_properties.get("collection_thumbnail") is not None
                or asset_properties.get("collection_overview") is not None
            ):
                thumbnail = Thumbnails()
                thumbnail.collection(
                    asset_properties["collection_thumbnail"],
                    asset_properties["collection_overview"],
                    dict(harvesting_vars)["services"],
                    dict(harvesting_vars)["dataset"],
                    dict(harvesting_vars),
                    dict(collection_dict)["collection_id"],
                    dict(harvesting_vars)["main_dataset_url"],
                    stac_catalog,
                    str(asset_properties.get("collection_thumbnail_link")),
                    str(asset_properties.get("collection_overview_link")),
                    logger_properties=self.logger_properties,
                )
            if asset_properties.get("collection_custom_asset") is not None:
                collection: pystac.Collection = constants.empty_collection
                collection = stac_catalog.get_child(dict(collection_dict)["collection_id"])  # type: ignore
                for asset in asset_properties["collection_custom_asset"]:
                    collection.add_asset(
                        key=asset.get("key"),
                        asset=pystac.Asset(
                            href=asset.get("href"),
                            title=asset.get("title")
                            if asset.get("title") is not None
                            else asset.get("key"),
                            roles=asset.get("roles")
                            if asset.get("roles") is not None
                            else ["data"],
                            media_type=asset.get("media_type")
                            if asset.get("media_type") is not None
                            else pystac.MediaType.TEXT,
                        ),
                    )



# SPDX-FileCopyrightText: 2023 Karlsruher Institut für Technologie
#
# SPDX-License-Identifier: CC0-1.0

import os
from typing import Union

from ..logger import Logger


class ExistenceValidator(object):
    """
    A class for verifying the main STAC catalog's existence.
    This class is implemented in :class:`~tds2stac.STACCreator`.

    Args:
        stac_dir (st, Optional): Directory of the main STAC catalog (*)
        logger_properties (dict, optional): A dictionary of properties for logger. default is `None`.

    """

    stac_dir: str
    """
    Directory of the main STAC catalog. It can be a relative or absolute path.
    """
    logger_properties: Union[dict, None]
    """
    A dictionary of properties for logger. default is `None`.
    You can look at keys in :class:`~tds2stac.logger.Logger` class.
    """

    def __init__(
        self,
        stac_dir: str = os.getcwd(),
        logger_properties: Union[dict, None] = dict(),
    ):
        self.stac_dir = stac_dir
        self.stac_dir = os.path.join(self.stac_dir, "/catalog.json")
        if os.path.exists(self.stac_dir):
            self.existence = True
        else:
            self.existence = False
        if logger_properties is not None:
            logger_properties["logger_msg"] = self.existence
        Logger(logger_properties)

    def __repr__(self):
        return "<TDS2STACExistenceChecker existence: %s>" % (self.existence)



import json
import sys
from datetime import datetime

import pystac
import pytz

from ..logger import Logger

# SPDX-FileCopyrightText: 2023 Karlsruher Institut für Technologie
#
# SPDX-License-Identifier: CC0-1.0
from ..statics import constants


class ExtraMetadata:
    """
    A class to add extra metadata to the STAC items and collections.
    Args:
        logger_properties (dict): A dictionary containing the logger properties.
    """

    def __init__(self, logger_properties: dict = dict()):
        self.logger_properties = logger_properties

    def item(
        self,
        item: pystac.Item,
        extra_metadata: dict = dict(),
        harvesting_vars: dict = dict(),
    ):
        """
        Add extra metadata to the STAC item.
        """
        try:
            start_datetime = datetime.strptime(
                harvesting_vars["temporal_extent_start_datetime"],
                "%Y-%m-%dT%H:%M:%S.%fZ",
            ).replace(tzinfo=pytz.utc)
        except ValueError:
            start_datetime = datetime.strptime(
                harvesting_vars["temporal_extent_start_datetime"],
                "%Y-%m-%dT%H:%M:%SZ",
            ).replace(tzinfo=pytz.utc)
        try:
            end_datetime = datetime.strptime(
                harvesting_vars["temporal_extent_end_datetime"],
                "%Y-%m-%dT%H:%M:%S.%fZ",
            ).replace(tzinfo=pytz.utc)
        except ValueError:
            end_datetime = datetime.strptime(
                harvesting_vars["temporal_extent_end_datetime"],
                "%Y-%m-%dT%H:%M:%SZ",
            ).replace(tzinfo=pytz.utc)
        if extra_metadata.get("extra_metadata_file") is not None:
            extra_metadata_json = self.open_json_file(
                str(extra_metadata.get("extra_metadata_file"))
            )
            if (
                extra_metadata_json != {}
                and extra_metadata_json.get("item") is not None
            ):
                for key, value in extra_metadata_json["item"].items():
                    if key == "properties":
                        item.properties.update(value)
                    if key == "extra_fields":
                        item.extra_fields.update(value)
                    if key == "providers":
                        item.common_metadata.providers = value
                    if key == "license":
                        item.common_metadata.license = value
                    if key == "constellation":
                        item.common_metadata.constellation = value
                    if key == "platform":
                        item.common_metadata.platform = value
                    if key == "instruments":
                        item.common_metadata.instruments = value
                    if key == "gsd":
                        item.common_metadata.gsd = value
                    if key == "providers":
                        item.common_metadata.providers = value
                    if key == "title":
                        item.common_metadata.title = value
                    if key == "description":
                        item.common_metadata.description = value
                    if key == "start_datetime":
                        item.common_metadata.start_datetime = start_datetime
                    if key == "end_datetime":
                        item.common_metadata.end_datetime = end_datetime
                    if key == "updated":
                        item.common_metadata.updated = end_datetime
                    if key == "created":
                        item.common_metadata.created = start_datetime

            else:
                self.logger_properties["logger_level"] = "WARNING"
                self.logger_properties[
                    "logger_msg"
                ] = "The `extra_metadata_file` is empty. So, the default extra metadata won't be added to the STAC item."
                Logger(self.logger_properties)
        else:
            self.logger_properties["logger_level"] = "WARNING"
            self.logger_properties[
                "logger_msg"
            ] = "The `extra_metadata_file` is not provided. So, the default extra metadata will be added to the STAC item."
            Logger(self.logger_properties)
            extra_metadata[
                "extra_metadata_file"
            ] = constants.default_extra_metadata_file
            extra_metadata_json = self.open_json_file(
                str(extra_metadata.get("extra_metadata_file"))
            )
            if (
                extra_metadata_json != {}
                and extra_metadata_json.get("item") is not None
            ):
                for key, value in extra_metadata_json["item"].items():
                    if key == "properties":
                        item.properties.update(value)
                    if key == "extra_fields":
                        item.extra_fields.update(value)
                    if key == "providers":
                        item.common_metadata.providers = value
                    if key == "license":
                        item.common_metadata.license = value
                    if key == "constellation":
                        item.common_metadata.constellation = value
                    if key == "platform":
                        item.common_metadata.platform = value
                    if key == "instruments":
                        item.common_metadata.instruments = value
                    if key == "gsd":
                        item.common_metadata.gsd = value
                    if key == "providers":
                        item.common_metadata.providers = value
                    if key == "title":
                        item.common_metadata.title = value
                    if key == "description":
                        item.common_metadata.description = value
                    if key == "start_datetime":
                        item.common_metadata.start_datetime = start_datetime
                    if key == "end_datetime":
                        item.common_metadata.end_datetime = end_datetime
                    if key == "updated":
                        item.common_metadata.updated = end_datetime
                    if key == "created":
                        item.common_metadata.created = start_datetime
            else:
                self.logger_properties["logger_level"] = "WARNING"
                self.logger_properties[
                    "logger_msg"
                ] = "The default `extra_metadata_file` is empty. So, the default extra metadata won't be added to the STAC item."
                Logger(self.logger_properties)

    def collection(
        self, collection: pystac.Collection, extra_metadata: dict = dict()
    ):
        """
        Add extra metadata to the STAC collection.
        """
        if extra_metadata.get("extra_metadata_file") is not None:
            extra_metadata_json = self.open_json_file(
                str(extra_metadata.get("extra_metadata_file"))
            )
            if (
                extra_metadata_json != {}
                and extra_metadata_json.get("collection") is not None
            ):
                for key, value in extra_metadata_json["collection"].items():
                    if key == "extra_fields":
                        collection.extra_fields.update(value)
                    if key == "keywords":
                        collection.keywords = value
                    if key == "providers":
                        collection.providers = value
                    if key == "license":
                        collection.license = value
            else:
                self.logger_properties["logger_level"] = "WARNING"
                self.logger_properties[
                    "logger_msg"
                ] = "The `extra_metadata_file` is empty. So, the default extra metadata won't be added to the STAC collection."
                Logger(self.logger_properties)
        else:
            self.logger_properties["logger_level"] = "WARNING"
            self.logger_properties[
                "logger_msg"
            ] = "The `extra_metadata_file` is not provided. So, the default extra metadata will be added to the STAC collection."
            Logger(self.logger_properties)
            extra_metadata[
                "extra_metadata_file"
            ] = constants.default_extra_metadata_file
            extra_metadata_json = self.open_json_file(
                str(extra_metadata.get("extra_metadata_file"))
            )
            if (
                extra_metadata_json != {}
                and extra_metadata_json.get("collection") is not None
            ):
                for key, value in extra_metadata_json["collection"].items():
                    if key == "extra_fields":
                        collection.extra_fields.update(value)
                    if key == "keywords":
                        collection.keywords = value
                    if key == "providers":
                        collection.providers = value
                    if key == "license":
                        collection.license = value
            else:
                self.logger_properties["logger_level"] = "WARNING"
                self.logger_properties[
                    "logger_msg"
                ] = "The default `extra_metadata_file` is empty. So, the default extra metadata won't be added to the STAC collection."
                Logger(self.logger_properties)

    def open_json_file(self, file_path: str) -> dict:
        """
        Open the JSON file.
        """
        try:
            loaded_json = json.load(open(file_path))
            return loaded_json

        except Exception:
            ex_type, ex_value, ex_traceback = sys.exc_info()
            self.logger_properties["logger_level"] = "ERROR"
            if ex_type is not None and ex_value is not None:
                self.logger_properties[
                    "logger_msg"
                ] = f"Could not open {file_path}. %s: %s" % (
                    ex_type.__name__,
                    ex_value,
                )
            else:
                self.logger_properties[
                    "logger_msg"
                ] = f"Could not open {file_path}."
            Logger(self.logger_properties)
            return {}



# SPDX-FileCopyrightText: 2023 Karlsruher Institut für Technologie
#
# SPDX-License-Identifier: CC0-1.0
from tds2stac import WebServiceListScraper


def test_WebServiceListScraper():
    WebServiceListScraper(
        url="http://localhost:8080/thredds/catalog/catalogs/scenario5/test/catalog.html",
        logger_properties={"logger_handler": "StreamHandler"},
        requests_properties={"timeout": 10, "verify": False},
    )



import os

# SPDX-FileCopyrightText: 2023 Karlsruher Institut für Technologie
#
# SPDX-License-Identifier: CC0-1.0
import sys
import traceback
from datetime import datetime
from typing import Literal, Union

import pytz
import urllib3
from lxml import etree
from tqdm import tqdm

from . import assets, creator, harvester, logger, utils
from .analysers.nested_collections import NestedCollectionInspector
from .analysers.properties_verifier import Verifier
from .analysers.recognizer import Recognizer
from .statics import constants
from .webservices import core

##################################################
# Disabling the warning of InsecureRequestWarning
# for web server that doesn't have SSL certificate
##################################################
urllib3.disable_warnings(urllib3.exceptions.InsecureRequestWarning)


class TDS2STACIntegrator(object):
    """
    This class is the central component of the TDS2STAC. It harvests the TDS catalog
    and then generates the STAC-Catalog, -Collections, and -Items through the TDS catalogs,
    based on the user's input.
    This class mainly defines all configurations related to harvesting and STAC creation.
    In the first step, it recognizes the scenario of the TDS catalog using :class:`~tds2stac.Recognizer`.
    If it is recognized as a nested collection, :class:`~tds2stac.NestedCollectionInspector` is responsible for
    determining the nested collection's `ID`, `Title`, and `url` of subdirectories. Other procedures follow
    in succession. For example, :class:`~tds2stac.CollectionHarvester` harvests the collection's information
    and :class:`~tds2stac.STACCreator` creates the STAC-Catalog and -Collection. Then, :class:`~tds2stac.ItemHarvester`
    harvests the item's information and :class:`~tds2stac.STACCreator` creates the STAC-Item and connect them
    to the related STAC-Collections. At the end each STAC-Collection will be connected to the main STAC-Catalog.

    Args:
        TDS_catalog(str) : The URL address of the TDS catalog that will be harvested.
        stac_dir (str, Optional) : Directory of saving created STAC catalogs.
        stac_id (str, Optional): STAC catalog ID. default value is 'TDS2STAC'.
        stac_title (str, optional): STAC catalog Title. default value is 'TDS2STAC'.
        stac_description (str, optional): STAC catalog description.
        stac_existence (bool, optional): Verifying the presence of the STAC catalog in order to
            update an existing catalog; if not, a new catalog will be generated.
        stac_existence_collection (bool, optional): Verifying the presence of the STAC Collection
            in order to update an existing catalog; if not, a new collection will be generated.
        collection_tuples (list, optional): The elements of this tuple comprise the auto-TDS2STAC-generated ID
            , the user-defined ID, title, and description of the STAC-Collection respectively. (auto-ID, user-ID,
            user-title, user-description).
        datetime_filter (list, optional): Datetime-based filtering of harvesting. It works based on the `modified`
            tag in each dataset at TDS.
        aggregated_dataset_url (str, optional): Dataset's URL of each data entry in the Aggregated datasets of TDS.
        depth_number (int, optional) : depth number of nested datasets if it is a nested collection. default value is 0.
        limited_number (int, optional): The objective is to reduce the quantity of harvested items in each collection. It is beneficial
            for developing and testing purposes.
        spatial_information (list, optional): Spatial information of 2D datasets e.g. [minx, maxx, miny, maxy] or 1D dataset e.g.
            [x,y]. default value is None.
        temporal_format_by_dataname (str, optional): A preferred datetime format for datasets that include the time period in their names. e.g
            "e%y%m%d%H.%M%S%f"
        item_geometry_linestring (bool, optional): Set True to make a LineString geometry for STAC-Items from wms service.
            Otherwise it makes Polygon geometry for the given Item. default value is False.
        extension_properties (dict, optional): A dictionary of properties for extensions. default is None.
            For more information about the keys, please refer to the :class:`~tds2stac.TDS2STACIntegrator.extension_properties`.
        webservice_properties (dict, optional): A dictionary of properties for web_service. default is None (optional)
            For more information about the keys, please refer to the :class:`~tds2stac.TDS2STACIntegrator.webservice_properties`.
        asset_properties (dict, optional): A dictionary of properties for assets. default is None (optional)
            For more information about the keys, please refer to the :class:`~tds2stac.TDS2STACIntegrator.asset_properties`.
        logger_properties (dict, optional): A dictionary of properties for logger. default is `None`.
        requests_properties (dict, optional): A dictionary that modify the requests to
            URLs. To obtain additional information on this topic, refer to
            the :class:`~tds2stac.TDS2STAC.requests_properties`. The default value is
            an empty dictionary.
       extra_metadata (dict, optional): A dictionary of extra metadata that
            you desire to include in the STAC-Collection and STAC-Items. To obtain
            additional information on this topic, please refer to the
            :class:`~tds2stac.TDS2STAC.extra_metadata`. The default value is an empty
            dictionary


    """

    TDS_catalog: str
    """
        TDS catalog URL address. Initial point of harvesting e.g.
        https://thredds.atmohub.kit.edu/thredds/catalog/caribic/IAGOS-CARIBIC_MS_files_collection_20231017/catalog.html
    """
    stac_dir: str
    """
    Directory of saving created STAC catalogs e.g. /path/to/stac/directory/
    """
    stac_id: str
    """
    STAC catalog ID. default value `TDS2STAC`
    """
    stac_title: Union[str, None]
    """
    STAC catalog Title. default value `TDS2STAC`
    """
    stac_description: Union[str, None]
    """
    STAC catalog description
    """
    stac_existence: Literal[False]
    """
    Verifying the existence of STAC catalog.
    If the catalog exists in the directory, it updates a existed catalog,
    otherwise it creates new catalog. default value `False`
    """
    stac_existence_collection: Literal[False]
    """
    Verifying the existence of STAC Collections.
    If the collection exists in the directory, it updates a existed collection,
    otherwise it creates new collection. default value `False`
    """
    collection_tuples: Union[list[tuple], None]
    """
    STAC collection auto-generated ID, user-ID, user-Title and user-Description defined by user.
    It is worth mentioning that in order to obtain the list of automatically generated collection
    IDs, one can employ the :class:`~tds2stac.NestedCollectionInspector` for the given
    TDS Catalog and subsequently utilize this argument.
    Warning - Identifiers should consist of only lowercase characters, numbers, '_', and '-'.
    Default value `None`. e.g. (ID, Title, Description)
    """
    datetime_filter: Union[list, None]
    """
    Datetime-based filtering. e.g. ``['2010-02-18T00:00:00.000Z','2020-02-22T00:00:00.000Z']``
    Default value `None`. It should be noted it works based on the `modified` tag in each dataset at TDS.
    """
    aggregated_dataset_url: Union[str, None]
    """
    Dataset's URL of each data entry in the Aggregated datasets of TDS.. default value `None`.
    The `HTTPServer` is not functional in the aggregated dataset. Therefore, in order to utilize
    this service as an asset in our STAC-Item, we should employ the `aggregated_dataset_url`, which
    links the individual datasets to the `HTTPServer` asset of the relevant Item.
    """
    depth_number: Union[int, None]
    """
    The depth refers to the number of layered datasets. If the collection is nested, this argument
    is applicable; otherwise, employing this argument would be futile. default value `None` (optional)
    """
    limited_number: Union[int, None]
    """
    The objective is to reduce the quantity of harvested items in each collection. It is beneficial
    for developing and testing purposes.. default value `None` (optional)
    """
    spatial_information: Union[list, None]
    """
    Spatial information of 2D datasets e.g. [minx, maxx, miny, maxy] or 1D dataset e.g. [x,y].
    Default value `None`(optional)
    """
    temporal_format_by_dataname: Union[str, None]
    """
    A preferred datetime format for datasets that include the time period in their names e.g "e%y%m%d%H.%M%S%f".
    Default value `None` (optional)
    """
    item_geometry_linestring: Literal[False]
    """
    The default value for the LineString geometry in the STAC Items from the WMS
    service is set to False and the default geometry type for the STAC-Item is Polygon.
    However, in instances where the item has a POINT geometry, it can be automatically detected.
    However, in order to obtain the LineString geometry, it is necessary to set this argument
    to True.
    """
    extension_properties: Union[dict, None]
    """
    A dictionary of properties for extensions. default is `None`.

        **item_extensions (list[str, tuple], optional)**:
            The argument can consist of either a list of extension names (string) or a
            list of tuples containing three elements: the extension name, the function
            name or class name associated with the extension, and the Python script
            required for execution. For more explanation, refer to the :ref:`custom-extension`.

        **collection_extensions (Union[list, tuple], optional)**:
            It works as same as `item_extensions` argument. For more explanation,
            refer to the :ref:`custom-extension`.

    """

    webservice_properties: Union[dict, None]
    """
    A dictionary of properties for web_service. default is `None`.

    It has the following keys.
        **web_service_config_file(str, opntional)**:
            The primary `tag_config.json` file is situated in the primary directory
            of the installed TDS2STAC. However, the user has the ability to declare
            an alternative `tag_config.json` file, which allows for customization of
            the settings. The user can specify the location of their own JSON file
            in this section. To obtain further details on the creation of a
            `tag_config.json` file, refer: :ref:`tag-config`. The default
            value is set to `tag_config.json` in the root directory of the installed
            app.
    """
    asset_properties: Union[dict, None]
    """
    A dictionary of properties for assets. default is `None`.
    When it's None, keys look like the following example:
        **collection_thumbnail (str, optional)**:
            A `thumbnail` asset for STAC-collection sourced from the Web Map Service (WMS)
            of the TDS. It can be chosen from `wms`, `link`, or `None`. The default value
            is set to None.
        **collection_overview (str, optional)**:
            A `overview` asset for STAC-collection sourced from the Web Map Service (WMS)
            of the TDS. It can be chosen from `wms`, `link`, or `None`. The default value
            is set to None.
        **collection_thumbnail_link (str, optional)**:
            This property is reliant upon the values of `collection_thumbnail` and
            `collection_overview`. When the value of either of these attributes is
            set to `link`, it allows for the inclusion of a hyperlink to an image
            for `collection_thumbnail` or `collection_overview`.
        **collection_overview_link (str, optional)**:
            This property is reliant upon the values of `collection_thumbnail` and
            `collection_overview`. When the value of either of these attributes is
            set to `link`, it allows for the inclusion of a hyperlink to an image
            for `collection_thumbnail` or `collection_overview`.
        **collection_custom_assets (list, optional)**:
            This is a list of asset dictionaris that includes the `key`, `href`, and
            `title`, `role` (as a list), and `media_type` of the asset. The default
            value is set to None. For more information, refer to the :ref:`custom-asset`.
        **item_thumbnail (bool, optional)**:
            A `thumbnail` asset for STAC-Items sourced from the Web Map Service (WMS)
            of the TDS. The default value is set to False.
        **item_overview (bool, optional)**:
            A `overview` asset for STAC-Items sourced from the Web Map Service (WMS)
            of the TDS. The default value is set to False.
        **item_getminmax_thumbnail (bool, optional)**:
            The TDS offers a function that allows users to obtain the minimum and maximum
            values of the colorbar associated with an image through the use of `metadata`.
            The aforementioned attribute is contingent upon both the `item_thumbnail` and
            `item_overview`. The default value is set to False.
        **assets_list_allowed (list, optional)**:
            This is a list of permissible web services that will be incorporated as assets
            in the STAC-Item. The :class:`~tds2stac.WebServiceScraper` class provides access
            to the list of available web services. Default value is None.
        **assets_list_avoided (list, optional)**:
            This is a list of web services that will be excluded from the STAC-Item asset
            list. The :class:`~tds2stac.WebServiceScraper` class provides access to the
            list of available webservices. Default value is None.
        **explore_data (bool, optional)**:
            By enabling the `True` setting, the inclusion of Godiva3 as an exploration asset will be implemented.
        **verify_explore_data (bool, optional)**:
            This argument verifies the availability of the `GetMetadata` function. The provided
            function facilitates the retrieval of data necessary for generating maps using the
            Web Map Service (WMS) protocol. However, an error occurs when attempting to open
            `Godiva3` when this function doesn't work. In order to mitigate such errors, it
            would be advisable to establish this argument.
        **jupyter_notebook (bool, optional)**:
            This argument posits the inclusion of the Jupyter Notebook as an asset.
    """
    logger_properties: dict
    """
    A dictionary of properties for logger. default is `None`.
    You can look at keys in :class:`~tds2stac.logger.Logger` class.
    """
    requests_properties: dict
    """
    A dictionary of properties that adjust the requests to URLs. It contains the following keys:

        **verify (bool, optional)**:
            It is a boolean that if it is True, it verifies the SSL certificate. By default it is False.
        **timeout (int, optional)**:
            It is an integer that sets the timeout of the requests. By default it is 10 seconds.
        **auth (tuple, optional)**:
            It is a tuple that contains the username and password for the authentication. By default it is None.
    """
    extra_metadata: dict
    """
    A dictionary of extra metadata that you want to add to the
    STAC-Collection and STAC-Items. It has two main keys,
    `extra_metadata` that is boolean and `extra_metadata_file`
    that is the address of `extra_metadata.json` JSON file. For
    getting more information about making the `extra_metadata.json`
    file, please refer to :ref:`extra_metadata`.
    By default, if 'extra_metadata' is set to True, the
    'extra_metadata.json' file is utilised for the 'extra_metadata_file'
    key, which is situated in the'sta2stac' main directory.
    """

    def __init__(
        self,
        TDS_catalog: str,
        stac_dir: str = os.getcwd(),
        # web_service: str,
        stac_id: str = "TDS2STAC",
        stac_title: Union[str, None] = "TDS2STAC",
        stac_description: Union[str, None] = None,
        stac_existence: bool = False,
        stac_existence_collection: bool = False,
        collection_tuples: Union[list[tuple], None] = None,
        datetime_filter: Union[list, None] = None,
        aggregated_dataset_url: Union[str, None] = None,
        depth_number: Union[int, None] = None,
        limited_number: Union[int, None] = None,
        # spatial_information :Union[list[Union[str, int], Union[str, int]]|list[Union[str, int], Union[str, int],Union[str, int],Union[str, int]], None] = None,
        spatial_information: Union[list, None] = None,
        temporal_format_by_dataname: Union[str, None] = None,
        item_geometry_linestring: bool = False,
        webservice_properties: Union[dict, None] = dict(),
        asset_properties: Union[dict, None] = dict(),
        extension_properties: Union[dict, None] = dict(),
        logger_properties: dict = dict(),
        requests_properties: dict = dict(),
        extra_metadata: dict = dict(),
    ) -> None:
        ################################################
        # Verifying the `webservice_properties`,
        # `asset_properties`, and `extension_properties`,
        # when they are None
        ################################################
        verifier = Verifier()
        if webservice_properties is not None and isinstance(
            webservice_properties, dict
        ):
            verifier.webservice_properties(webservice_properties)
        if asset_properties is not None and isinstance(asset_properties, dict):
            verifier.asset_properties(asset_properties)
        if extension_properties is not None and isinstance(
            extension_properties, dict
        ):
            verifier.extension_properties(extension_properties)
        if logger_properties is not None and isinstance(
            logger_properties, dict
        ):
            verifier.logger_properties(logger_properties)
        if (
            requests_properties is not None
            and requests_properties != {}
            and isinstance(requests_properties, dict)
        ):
            verifier.requests_properties(requests_properties)
            self.requests_properties = requests_properties
        self.requests_properties = requests_properties

        if logger_properties is not None and isinstance(
            logger_properties, dict
        ):
            self.logger_properties = logger_properties
        if extra_metadata is not None and isinstance(extra_metadata, dict):
            verifier.extra_metadata(extra_metadata)

        self.logger_properties["logger_level"] = "DEBUG"
        self.logger_properties["logger_msg"] = "Harvesting is started!"
        logger.Logger(self.logger_properties)

        ################################################
        # Getting the list of used webservices names in
        # `tag_config.json`
        ################################################
        if webservice_properties is not None:
            if (
                webservice_properties.get("web_service_config_file")
                is not None
                and webservice_properties["web_service_config_file"]
                == "default"
            ):
                webservice_properties["webservices"] = list(
                    core.ConfigFileWebServicesScraper(
                        constants.default_tag_config_file,
                        self.logger_properties,
                    )
                )
            elif (
                webservice_properties["web_service_config_file"] is not None
                and webservice_properties["web_service_config_file"]
                != "default"
            ):
                webservice_properties["webservices"] = list(
                    core.ConfigFileWebServicesScraper(
                        webservice_properties["web_service_config_file"],
                        self.logger_properties,
                    )
                )
        ################################################
        # Defining a dict to store the items information
        ################################################
        item_harvested_details = None
        ################################################
        # Getting the Date time modiefied of TDS catalog
        ################################################
        if datetime_filter is not None:
            """Skip TDS datasets out of 'datetime_filter' according
            to 'modified' attribute in `date` tag"""

            if datetime_filter[0] is not None:
                try:
                    datetime_after = datetime.strptime(
                        datetime_filter[0], "%Y-%m-%dT%H:%M:%S.%fZ"
                    )

                    if not isinstance(datetime_after, datetime):
                        self.logger_properties[
                            "logger_msg"
                        ] = "`datetime_after` parameter have to be a datatime object. Therefore `datetime_filter` is not applied for this harvesting."
                        self.logger_properties["logger_level"] = "WARNING"
                        logger.Logger(self.logger_properties)

                    else:
                        if datetime_after.tzinfo:
                            datetime_after = datetime_after.astimezone(
                                pytz.utc
                            )
                        else:
                            datetime_after = datetime_after.replace(
                                tzinfo=pytz.utc
                            )
                except Exception:
                    datetime_after = None
                    ex_type, ex_value, ex_traceback = sys.exc_info()
                    if ex_type is not None and ex_value is not None:
                        self.logger_properties[
                            "logger_msg"
                        ] = "`datetime_filter` warning: %s : %s" % (
                            ex_type.__name__,
                            ex_value,
                        )
                        self.logger_properties["logger_level"] = "WARNING"
                        logger.Logger(self.logger_properties)
                        print(traceback.format_exc())

            if datetime_filter[1] is not None:
                try:
                    datetime_before = datetime.strptime(
                        datetime_filter[1], "%Y-%m-%dT%H:%M:%S.%fZ"
                    )

                    if not isinstance(datetime_before, datetime):
                        self.logger_properties[
                            "logger_msg"
                        ] = "`datetime_before` parameter have to be a datatime object. Therefore `datetime_filter` is not applied for this harvesting."
                        self.logger_properties["logger_level"] = "WARNING"
                        logger.Logger(self.logger_properties)
                    else:
                        if datetime_before.tzinfo:
                            datetime_before = datetime_before.astimezone(
                                pytz.utc
                            )
                        else:
                            datetime_before = datetime_before.replace(
                                tzinfo=pytz.utc
                            )
                except Exception:
                    datetime_before = None
                    if ex_type is not None and ex_value is not None:
                        self.logger_properties[
                            "logger_msg"
                        ] = "`datetime_filter` warning: %s : %s" % (
                            ex_type.__name__,
                            ex_value,
                        )
                        self.logger_properties["logger_level"] = "WARNING"
                        logger.Logger(self.logger_properties)
                    print(traceback.format_exc())
        else:
            datetime_after = None
            datetime_before = None
        ###############################
        # Finding dataset's scenario
        ###############################
        recognizer = Recognizer(
            main_catalog_url=TDS_catalog,
            logger_properties=self.logger_properties,
            requests_properties=self.requests_properties,
        )
        ###############################
        # STAC-Catalog creation
        ###############################
        STAC_creator = creator.STACCreator()
        stac_catalog = STAC_creator.STACCatalog(
            url=TDS_catalog,
            stac_id=stac_id,
            stac_title=stac_title,
            stac_desc=stac_description,
            stac_dir=stac_dir,
            stac_existence=stac_existence,
            logger_properties=self.logger_properties,
            requests_properties=self.requests_properties,
        )
        validate_value = utils.validate_catalog_url(
            url=TDS_catalog, requests_properties=self.requests_properties
        )
        if not validate_value:
            self.logger_properties["logger_level"] = "CRITICAL"
            self.logger_properties[
                "logger_msg"
            ] = "The TDS catalog URL is not valid! Please check the URL and try again."
            logger.Logger(self.logger_properties)
            return
        else:
            self.logger_properties["logger_level"] = "DEBUG"
            self.logger_properties[
                "logger_msg"
            ] = "The TDS catalog URL is valid!"
            logger.Logger(self.logger_properties)
        ########################################
        # STAC-Collection and -Item creation
        # 1. Harvest for nested collections
        ########################################
        if any(
            scenario in str(recognizer.status)
            for scenario in constants.nested_scenarios
        ):
            list_of_collection_details = NestedCollectionInspector(
                main_catalog_url=TDS_catalog,
                nested_number=depth_number,
                logger_properties=self.logger_properties,
                requests_properties=self.requests_properties,
            )
            for k in list_of_collection_details:
                ###########################################
                # defining a new function that harvests the
                # collections and create STAC-Collections
                # - Harvest Collection's information
                ###########################################
                collection_dict = harvester.CollectionHarvester(
                    TDS_catalog,
                    recognizer.status,
                    subdirs=k,
                    collection_tuples=collection_tuples,
                    logger_properties=self.logger_properties,
                    requests_properties=self.requests_properties,
                )
                ###########################
                # - Create STAC-Collection
                ###########################
                stac_collection = STAC_creator.STACCollection(
                    catalog=stac_catalog,
                    collection_id=dict(collection_dict)["collection_id"],
                    collection_title=dict(collection_dict)["collection_title"],
                    collection_description=dict(collection_dict)[
                        "collection_description"
                    ],
                    stac_existence_collection=stac_existence_collection,
                    extra_metadata=extra_metadata,
                )
                ###########################################
                # - Loop over collection's items
                ###########################################
                if k[3] != []:
                    for url_for_items in k[3]:
                        (
                            xml_url_catalog,
                            id_catalog,
                            xml,
                        ) = utils.xml_processing(
                            url_for_items, self.requests_properties
                        )
                        recognizer_output = Recognizer(
                            main_catalog_url=xml_url_catalog,
                            logger_properties=self.logger_properties,
                            requests_properties=self.requests_properties,
                        )

                        try:
                            tree = etree.XML(xml)
                        except BaseException:
                            continue
                        data_counted: int = 0

                        for data in tqdm(
                            tree.findall(
                                ".//{%s}dataset[@urlPath]" % constants.unidata
                            ),
                            colour="red",
                        ):
                            data_counted += 1
                            if limited_number is not None:
                                if data_counted > limited_number:
                                    break
                            if (
                                utils.replacement_func_collection_item_id(
                                    utils.replacement_func(data.get("ID"))
                                )
                                in stac_collection["existed_items_id_list"]
                            ):
                                continue
                            self.harvesting_vars = constants.harvesting_vars
                            self.harvesting_vars[
                                "collection_interval_time"
                            ] = (  # noqa
                                dict(item_harvested_details)["collection_interval_time"] if item_harvested_details is not None else []  # type: ignore
                            )
                            self.harvesting_vars[
                                "collection_interval_time_final"
                            ] = (  # noqa
                                dict(item_harvested_details)["collection_interval_time_final"] if item_harvested_details is not None else []  # type: ignore
                            )
                            self.harvesting_vars["collection_bbox"] = (  # noqa
                                dict(item_harvested_details)["collection_bbox"] if item_harvested_details is not None else []  # type: ignore
                            )
                            self.harvesting_vars[
                                "collection_footprint"
                            ] = (  # noqa
                                dict(item_harvested_details)["collection_footprint"] if item_harvested_details is not None else None  # type: ignore
                            )
                            self.harvesting_vars["collection_footprint_point"] = dict(item_harvested_details)["collection_footprint_point"] if item_harvested_details is not None else None  # type: ignore
                            ################################
                            # - Harvest Item
                            #################################
                            item_harvested_details = harvester.ItemHarvester(
                                xml_url_catalog,
                                data,
                                self.harvesting_vars,
                                webservice_properties,
                                datetime_after=datetime_after,
                                datetime_before=datetime_before,
                                spatial_information=spatial_information,
                                temporal_format_by_dataname=temporal_format_by_dataname,
                                extension_properties=extension_properties,
                                linestring=item_geometry_linestring,
                                requests_properties=self.requests_properties,
                                logger_properties=self.logger_properties,
                            )

                            ################################
                            # - Create STAC-Item
                            #################################

                            STAC_creator.STACItem(
                                xml_url_catalog,
                                stac_catalog,
                                dict(item_harvested_details),
                                recognizer_output.status,
                                dict(collection_dict)["collection_id"],
                                aggregated_dataset_url=aggregated_dataset_url,
                                extension_properties=extension_properties,
                                asset_properties=asset_properties,
                                extra_metadata=extra_metadata,
                                stac_existence_collection=stac_existence_collection,
                                collection_bbox_existed=dict(stac_collection)[
                                    "collection_bbox_existed"
                                ],
                                collection_interval_time_final_existed=dict(
                                    stac_collection
                                )["collection_interval_time_final_existed"],
                                logger_properties=self.logger_properties,
                            )
                            if asset_properties is not None and (
                                asset_properties["collection_thumbnail"]
                                in [
                                    "wms",
                                    "link",
                                ]
                                or asset_properties["collection_overview"]
                                in ["wms", "link"]
                                or asset_properties.get(
                                    "collection_custom_asset"
                                )
                                is not None
                            ):
                                if (item_harvested_details) is not None:
                                    asset = assets.Assets()
                                    asset.collection(
                                        harvesting_vars=dict(
                                            item_harvested_details
                                        ),
                                        collection_dict=dict(collection_dict),
                                        stac_catalog=stac_catalog,
                                        asset_properties=asset_properties,
                                        logger_properties=self.logger_properties,
                                    )

                else:
                    (
                        xml_url_catalog,
                        id_catalog,
                        xml,
                    ) = utils.xml_processing(k[0], self.requests_properties)
                    recognizer_output = Recognizer(
                        main_catalog_url=xml_url_catalog,
                        logger_properties=self.logger_properties,
                        requests_properties=self.requests_properties,
                    )

                    try:
                        tree = etree.XML(xml)
                    except BaseException:
                        return
                    data_counted = 0
                    for data in tqdm(
                        tree.findall(
                            ".//{%s}dataset[@urlPath]" % constants.unidata
                        ),
                        colour="red",
                    ):
                        data_counted += 1
                        if limited_number is not None:
                            if data_counted > limited_number:
                                break
                        if (
                            utils.replacement_func_collection_item_id(
                                utils.replacement_func(data.get("ID"))
                            )
                            in stac_collection["existed_items_id_list"]
                        ):
                            continue

                        harvesting_vars = {  # type: ignore
                            "item_id": None,
                            "description": None,
                            "horizontal_ids_lat": None,
                            "horizontal_ids_lon": None,
                            "horizontal_axis_x": None,
                            "horizontal_axis_y": None,
                            "horizontal_extent_lon_min": None,
                            "horizontal_extent_lon_max": None,
                            "horizontal_extent_lat_min": None,
                            "horizontal_extent_lat_max": None,
                            "horizontal_description_lon": None,
                            "horizontal_description_lat": None,
                            "horizontal_reference_system": None,
                            "vertical_ids": None,
                            "vertical_axis": None,
                            "vertical_extent_upper": None,
                            "vertical_extent_lower": None,
                            "vertical_description": None,
                            "temporal_id": None,
                            "temporal_axis": None,
                            "temporal_extent_start_datetime": None,
                            "temporal_extent_end_datetime": None,
                            "temporal_description": None,
                            "variable_description": None,  # Description of each variable
                            "variable_dimensions": None,  # dimension of each variable
                            "variable_ids": None,  # Variable names
                            "variable_unit": None,  # Variable units
                            "variable_types": None,  # Variable types
                            "services": None,
                            "dataset": None,
                            "catalog_url": None,
                            "main_dataset_url": None,
                            "catalog_id": None,
                            "item_bbox": None,
                            "item_footprint": None,
                            "collection_bbox": None,
                            "collection_footprint": None,
                            "collection_footprint_point": None,
                            "collection_interval_time": None,
                            "modified_date_time": None,
                            "collection_interval_time_final": None,
                        }

                        harvesting_vars["collection_interval_time"] = (  # noqa
                            dict(item_harvested_details)["collection_interval_time"] if item_harvested_details is not None else []  # type: ignore
                        )
                        harvesting_vars[
                            "collection_interval_time_final"
                        ] = (  # noqa
                            dict(item_harvested_details)["collection_interval_time_final"] if item_harvested_details is not None else []  # type: ignore
                        )
                        harvesting_vars["collection_bbox"] = (  # noqa
                            dict(item_harvested_details)["collection_bbox"] if item_harvested_details is not None else []  # type: ignore
                        )
                        harvesting_vars["collection_footprint"] = (  # noqa
                            dict(item_harvested_details)["collection_footprint"] if item_harvested_details is not None else None  # type: ignore
                        )
                        harvesting_vars[
                            "collection_footprint_point"
                        ] = (  # noqa
                            dict(item_harvested_details)["collection_footprint_point"] if item_harvested_details is not None else None  # type: ignore
                        )
                        ################################
                        # - Harvest Item
                        #################################

                        item_harvested_details = harvester.ItemHarvester(
                            xml_url_catalog,
                            data,
                            harvesting_vars,
                            webservice_properties,
                            datetime_after=datetime_after,
                            datetime_before=datetime_before,
                            spatial_information=spatial_information,
                            temporal_format_by_dataname=temporal_format_by_dataname,
                            extension_properties=extension_properties,
                            linestring=item_geometry_linestring,
                            requests_properties=self.requests_properties,
                            logger_properties=self.logger_properties,
                        )

                        ################################
                        # - Create STAC-Item
                        #################################
                        STAC_creator.STACItem(
                            xml_url_catalog,
                            stac_catalog,
                            dict(item_harvested_details),
                            recognizer_output.status,
                            dict(collection_dict)["collection_id"],
                            aggregated_dataset_url=aggregated_dataset_url,
                            extension_properties=extension_properties,
                            asset_properties=asset_properties,
                            extra_metadata=extra_metadata,
                            stac_existence_collection=stac_existence_collection,
                            collection_bbox_existed=dict(stac_collection)[
                                "collection_bbox_existed"
                            ],
                            collection_interval_time_final_existed=dict(
                                stac_collection
                            )["collection_interval_time_final_existed"],
                            logger_properties=self.logger_properties,
                        )

                    if asset_properties is not None and (
                        asset_properties["collection_thumbnail"]
                        in [
                            "wms",
                            "link",
                        ]
                        or asset_properties["collection_overview"]
                        in ["wms", "link"]
                        or asset_properties.get("collection_custom_asset")
                        is not None
                    ):
                        if (item_harvested_details) is not None:
                            asset = assets.Assets()
                            asset.collection(
                                harvesting_vars=dict(item_harvested_details),
                                collection_dict=dict(collection_dict),
                                stac_catalog=stac_catalog,
                                asset_properties=asset_properties,
                                logger_properties=self.logger_properties,
                            )

        ########################################
        # 1. Harvest for none-nested collections
        ########################################
        else:
            ###########################################
            # defining a new function that harvests the
            # collections and create STAC-Collections
            # - Harvest Collection
            ###########################################

            collection_dict = harvester.CollectionHarvester(
                TDS_catalog,
                recognizer.status,
                subdirs=None,
                collection_tuples=collection_tuples,
                logger_properties=self.logger_properties,
                requests_properties=self.requests_properties,
            )
            ###########################
            # - Create STAC-Collection
            ###########################
            stac_collection = STAC_creator.STACCollection(
                catalog=stac_catalog,
                collection_id=dict(collection_dict)["collection_id"],
                collection_title=dict(collection_dict)["collection_title"],
                collection_description=dict(collection_dict)[
                    "collection_description"
                ],
                stac_existence_collection=stac_existence_collection,
                extra_metadata=extra_metadata,
            )
            ###########################################
            # - Loop over collection's items
            ###########################################
            (
                xml_url_catalog,
                id_catalog,
                xml,
            ) = utils.xml_processing(TDS_catalog, self.requests_properties)
            recognizer_output = Recognizer(
                main_catalog_url=xml_url_catalog,
                logger_properties=self.logger_properties,
                requests_properties=self.requests_properties,
            )
            try:
                tree = etree.XML(xml)
            except BaseException:
                return
            data_counted = 0

            for data in tqdm(
                tree.findall(".//{%s}dataset[@urlPath]" % constants.unidata),
                colour="red",
            ):
                data_counted += 1
                if limited_number is not None:
                    if data_counted > limited_number:
                        break
                # if data_counted % 50 == 0:
                #     import gc
                #     gc.collect()
                #     del stac_catalog
                #     del stac_collection
                #     stac_catalog = None
                #     stac_collection = None
                #     stac_catalog = stac_catalog_temp
                #     stac_collection = stac_collection_temp

                if (
                    utils.replacement_func_collection_item_id(
                        utils.replacement_func(data.get("ID"))
                    )
                    in stac_collection["existed_items_id_list"]
                ):
                    continue
                harvesting_vars = {  # type: ignore
                    "item_id": None,
                    "description": None,
                    "horizontal_ids_lat": None,
                    "horizontal_ids_lon": None,
                    "horizontal_axis_x": None,
                    "horizontal_axis_y": None,
                    "horizontal_extent_lon_min": None,
                    "horizontal_extent_lon_max": None,
                    "horizontal_extent_lat_min": None,
                    "horizontal_extent_lat_max": None,
                    "horizontal_description_lon": None,
                    "horizontal_description_lat": None,
                    "horizontal_reference_system": None,
                    "vertical_ids": None,
                    "vertical_axis": None,
                    "vertical_extent_upper": None,
                    "vertical_extent_lower": None,
                    "vertical_description": None,
                    "temporal_id": None,
                    "temporal_axis": None,
                    "temporal_extent_start_datetime": None,
                    "temporal_extent_end_datetime": None,
                    "temporal_description": None,
                    "variable_description": None,  # Description of each variable
                    "variable_dimensions": None,  # dimension of each variable
                    "variable_ids": None,  # Variable names
                    "variable_unit": None,  # Variable units
                    "variable_types": None,  # Variable types
                    "services": None,
                    "dataset": None,
                    "catalog_url": None,
                    "main_dataset_url": None,
                    "catalog_id": None,
                    "item_bbox": None,
                    "item_footprint": None,
                    "collection_bbox": None,
                    "collection_bbox_existed": None,
                    "collection_footprint": None,
                    "collection_footprint_point": None,
                    "collection_interval_time": None,
                    "modified_date_time": None,
                    "collection_interval_time_final": None,
                    "collection_interval_time_final_existed": None,
                }

                harvesting_vars["collection_interval_time"] = (  # noqa
                    dict(item_harvested_details)["collection_interval_time"] if item_harvested_details is not None else []  # type: ignore
                )
                harvesting_vars["collection_interval_time_final"] = (  # noqa
                    dict(item_harvested_details)["collection_interval_time_final"] if item_harvested_details is not None else []  # type: ignore
                )
                harvesting_vars["collection_bbox"] = (  # noqa
                    dict(item_harvested_details)["collection_bbox"] if item_harvested_details is not None else []  # type: ignore
                )
                harvesting_vars["collection_footprint"] = (  # noqa
                    dict(item_harvested_details)["collection_footprint"] if item_harvested_details is not None else None  # type: ignore
                )
                harvesting_vars["collection_footprint_point"] = (  # noqa
                    dict(item_harvested_details)["collection_footprint_point"] if item_harvested_details is not None else None  # type: ignore
                )
                ################################
                # - Harvest Item
                #################################

                item_harvested_details = harvester.ItemHarvester(
                    xml_url_catalog,
                    data,
                    harvesting_vars,
                    webservice_properties,
                    datetime_after=datetime_after,
                    datetime_before=datetime_before,
                    spatial_information=spatial_information,
                    temporal_format_by_dataname=temporal_format_by_dataname,
                    extension_properties=extension_properties,
                    linestring=item_geometry_linestring,
                    requests_properties=self.requests_properties,
                    logger_properties=self.logger_properties,
                )

                ################################
                # - Create STAC-Item
                #################################
                STAC_creator.STACItem(
                    xml_url_catalog,
                    stac_catalog,
                    dict(item_harvested_details),
                    recognizer_output.status,
                    dict(collection_dict)["collection_id"],
                    aggregated_dataset_url=aggregated_dataset_url,
                    extension_properties=extension_properties,
                    asset_properties=asset_properties,
                    extra_metadata=extra_metadata,
                    stac_existence_collection=stac_existence_collection,
                    collection_bbox_existed=dict(stac_collection)[
                        "collection_bbox_existed"
                    ],
                    collection_interval_time_final_existed=dict(
                        stac_collection
                    )["collection_interval_time_final_existed"],
                    logger_properties=self.logger_properties,
                )
                data.clear()
            if asset_properties is not None and (
                asset_properties["collection_thumbnail"]
                in [
                    "wms",
                    "link",
                ]
                or asset_properties["collection_overview"] in ["wms", "link"]
                or asset_properties.get("collection_custom_asset") is not None
            ):
                if (item_harvested_details) is not None:
                    asset = assets.Assets()
                    asset.collection(
                        harvesting_vars=dict(item_harvested_details),
                        collection_dict=dict(collection_dict),
                        stac_catalog=stac_catalog,
                        asset_properties=asset_properties,
                        logger_properties=self.logger_properties,
                    )

        #############################
        # Saving STAC catalog
        #############################
        self.logger_properties["logger_level"] = "DEBUG"
        self.logger_properties["logger_msg"] = "Harvesting is Finished!"
        logger.Logger(self.logger_properties)
        self.logger_properties["logger_level"] = "DEBUG"
        self.logger_properties[
            "logger_msg"
        ] = "Creating STAC-Metadata is started!"
        logger.Logger(self.logger_properties)

        save_out = STAC_creator.SaveCatalog(
            catalog=stac_catalog,
            catalog_dir=stac_dir,
            logger_properties=self.logger_properties,
        )
        if save_out:
            self.logger_properties["logger_level"] = "DEBUG"
            self.logger_properties[
                "logger_msg"
            ] = "Creating STAC-Metadata is finished!"
            logger.Logger(self.logger_properties)
        else:
            self.logger_properties["logger_level"] = "CRITICAL"
            self.logger_properties[
                "logger_msg"
            ] = "Creating STAC-Metadata is failed!"
            logger.Logger(self.logger_properties)
